# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["IdlMetadata", "TopicMetadata", "FieldMetadata", "parse_idl"]

import re


class IdlMetadata:
    """Information about SAL topics in an IDL file.

    Parameters
    ----------
    name : `str`
        SAL component name, e.g. "ATPtg".
    idl_path : `pathlib.Path`
        Path to IDL file.
    sal_version : `str` or `None`
        Version of ts_sal used to generate the IDL file,
        or `None` if not found.
    xml_version : `str`
        Version of ts_xml used to generate the IDL file,
        or `None` if not found.
    topic_info : `dict` [`str`: `TopicMetadata`]
        Dict of SAL topic name: topic metadata.
    """

    def __init__(self, name, idl_path, sal_version, xml_version, topic_info):
        self.name = name
        self.idl_path = idl_path
        self.sal_version = sal_version
        self.xml_version = xml_version
        self.topic_info = topic_info

    def __repr__(self):
        return (
            f"IdlMetadata(idl_path={self.idl_path}, "
            f"sal_version={repr(self.sal_version)}, "
            f"xml_version={repr(self.xml_version)})"
        )


class TopicMetadata:
    """Metadata about a topic.

    Parameters
    ----------
    sal_name : `str`
        SAL topic name, e.g. `logevent_summaryState`.
    version_hash : `str`
        Version has code.
    description : `str` or `None`
        Topic description, or `None` if unknown.

    Attributes
    ----------
    field_info : `dict` [`str`, `FieldMetadata`]
        Dict of field name: field metadata.
    """

    def __init__(self, sal_name, version_hash, description):
        self.sal_name = sal_name
        self.version_hash = version_hash
        self.description = description
        self.field_info = dict()

    def __repr__(self):
        return f"TopicMetadata(sal_name={repr(self.sal_name)}, description={self.description})"


class FieldMetadata:
    """Information about a field.

    Parameters
    ----------
    name : `str`
        Field name.
    description : `str` or `None`
        Description; `None` if not specified.
    units : `str` or `None`
        Units; `None` if not specified.
    type_name : `str`
        Data type name from the IDL file, e.g.
        "string<8>", "float", "double".
        This may not match the Python data type of the field.
        For instance dds maps most integer types to `int`,
        and both "float" and "double" to `double`.
    array_length : `int`
        Number of elements if an array; None if not an array.
    str_length : `int`
        Maximum allowed string length; None if unspecified (no limit)
        or not a string.
    """

    def __init__(self, name, description, units, type_name, array_length, str_length):
        self.name = name
        self.description = description
        self.units = units
        self.type_name = type_name
        self.array_length = array_length
        self.str_length = str_length

    def __repr__(self):
        return (
            f"FieldMetadata(name={repr(self.name)}, "
            f"description={repr(self.description)}, "
            f"units={repr(self.units)}, "
            f"type_name={repr(self.type_name)},"
            f"array_length={self.array_length}"
            f"str_length={self.str_length})"
        )

    def __str__(self):
        return f"description={repr(self.description)}, units={repr(self.units)}"


def parse_idl(name, idl_path):
    """Parse the SAL-generated IDL file.

    Parameters
    ----------
    name : str`
        SAL component name, e.g. "ATPtg".
    idl_path : `str`, `bytes`, or `pathlib.Path`
        Path to IDL file

    Returns
    -------
    metadata : `IdlMetadata`
        Parsed metadata.
    """
    # List of field types used in IDL, excluding "unsigned".
    type_names = (
        "string",
        "boolean",
        "float",
        "double",
        "octet",
        "short",
        "int",
        "long",
        "long long",
    )
    type_names_str = "|".join(type_names)

    sal_version_pattern = re.compile(
        r"//\sSAL_VERSION=(?P<sal_version>\S+) XML_VERSION=(?P<xml_version>\S+)"
    )
    topic_start_pattern = re.compile(
        r"\s*struct\s+(?P<sal_topic_name>.+)_(?P<version_hash>[a-zA-Z0-9]+) +{"
        r'\s*(?:// @Metadata=\((?:Description="(?P<description>.*)")?\))?'
    )
    topic_end_pattern = re.compile(r"\s};")
    field_pattern = re.compile(
        fr"\s*(?P<type_name>(?:unsigned )?(?:{type_names_str}))(?:<(?P<str_length>\d+)>)?"
        r"\s+(?P<field_name>[a-zA-Z][a-zA-Z0-9_-]*)(?:\[(?P<array_length>\d+)\])?;"
        r"\s*(?://\s*private)?"
        r'\s*(?:// @Metadata=\((?:Units="(?P<units>[^"]*)"),?'
        r'\s*(?:Description="(?P<description>.*)")?\))?'
    )

    # dict of sal_topic_name: `TopicMetadata`
    topic_info = dict()

    # Set topic_metadata to a `TopicMetadata` when parsing data for a topic
    # and None when not.
    topic_metadata = None
    isfirst = True
    sal_version = None
    xml_version = None
    with open(idl_path, "r") as f:
        for line in f:
            # The first line should contain version information
            # encoded as a comment.
            if isfirst:
                isfirst = False
                version_match = sal_version_pattern.match(line)
                if version_match:
                    sal_version = version_match.group("sal_version")
                    xml_version = version_match.group("xml_version")
                    continue

            if topic_metadata is None:
                # Look for the start of a new topic.
                topic_match = topic_start_pattern.match(line)
                if not topic_match:
                    continue

                sal_topic_name = topic_match.group("sal_topic_name")
                if sal_topic_name in topic_info:
                    raise RuntimeError(
                        f"Already parsed data for topic {sal_topic_name}"
                    )
                version_hash = topic_match.group("version_hash")
                topic_description = topic_match.group("description")
                topic_metadata = TopicMetadata(
                    sal_name=sal_topic_name,
                    version_hash=version_hash,
                    description=topic_description,
                )
                topic_info[sal_topic_name] = topic_metadata
            else:
                # Look for the next field or the end of this topic.
                field_match = field_pattern.match(line)
                if field_match:
                    field_name = field_match.group("field_name")
                    array_length = field_match.group("array_length")
                    if array_length is not None:
                        array_length = int(array_length)
                    str_length = field_match.group("str_length")
                    if str_length is not None:
                        str_length = int(str_length)
                    topic_metadata.field_info[field_name] = FieldMetadata(
                        name=field_name,
                        type_name=field_match.group("type_name"),
                        description=field_match.group("description"),
                        units=field_match.group("units"),
                        array_length=array_length,
                        str_length=str_length,
                    )
                elif topic_end_pattern.match(line):
                    topic_metadata = None

    return IdlMetadata(
        name=name,
        idl_path=idl_path,
        sal_version=sal_version,
        xml_version=xml_version,
        topic_info=topic_info,
    )



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import pathlib
import time
import unittest
import warnings

from lsst.ts import salobj

try:
    import SALPY_Test
except ImportError:
    SALPY_Test = None

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

INITIAL_LOG_LEVEL = 20
SAL__CMD_COMPLETE = 303

index_gen = salobj.index_generator()


class SALPYTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        self.datadir = pathlib.Path(__file__).resolve().parent / "data"
        self.index = next(index_gen)

    @unittest.skipIf(SALPY_Test is None, "Could not import SALPY_Test")
    async def test_salpy_remote_salobj_controller(self):
        await self.check_salpy_remote("minimal_salobj_controller.py")

    @unittest.skipIf(SALPY_Test is None, "Could not import SALPY_Test")
    async def test_salpy_remote_salpy_controller(self):
        await self.check_salpy_remote("minimal_salpy_controller.py")

    async def check_salpy_remote(self, exec_name):
        # Create at least the scalars telemetry reader before creating
        # the subprocess, to be sure the remote sees telemetry from the
        # subprocess (telemetry is volatile, so has no historical data).
        print(f"Remote: create SALPY object and topics with index={self.index}")
        t0 = time.monotonic()
        manager = SALPY_Test.SAL_Test(self.index)
        manager.setDebugLevel(0)
        manager.salEventSub("Test_logevent_logLevel")
        manager.salTelemetrySub("Test_scalars")
        manager.salCommand("Test_command_setLogLevel")
        dt = time.monotonic() - t0
        print(f"Remote: took {dt:0.2f} seconds to create topics")

        print(f"Remote: start {exec_name} in a subprocess")
        script_path = self.datadir / exec_name
        process = await asyncio.create_subprocess_exec(
            str(script_path), str(self.index), str(INITIAL_LOG_LEVEL)
        )

        try:

            async def get_logLevel():
                data = SALPY_Test.Test_logevent_logLevelC()
                while True:
                    retcode = manager.getEvent_logLevel(data)
                    if retcode == SALPY_Test.SAL__OK:
                        return data
                    elif retcode != SALPY_Test.SAL__NO_UPDATES:
                        raise RuntimeError(f"Unexpected return code {retcode}")
                    await asyncio.sleep(0.01)

            async def get_scalars():
                data = SALPY_Test.Test_scalarsC()
                while True:
                    retcode = manager.getNextSample_scalars(data)
                    if retcode == SALPY_Test.SAL__OK:
                        return data
                    elif retcode != SALPY_Test.SAL__NO_UPDATES:
                        raise RuntimeError(f"Unexpected return code {retcode}")
                    await asyncio.sleep(0.01)

            async def send_setLogLevel(level):
                done_ack_codes = frozenset(
                    (
                        SALPY_Test.SAL__CMD_ABORTED,
                        SALPY_Test.SAL__CMD_COMPLETE,
                        SALPY_Test.SAL__CMD_FAILED,
                        SALPY_Test.SAL__CMD_NOACK,
                        SALPY_Test.SAL__CMD_NOPERM,
                        SALPY_Test.SAL__CMD_STALLED,
                        SALPY_Test.SAL__CMD_TIMEOUT,
                    )
                )

                cmd_data = SALPY_Test.Test_command_setLogLevelC()
                cmd_data.level = level
                cmd_id = manager.issueCommand_setLogLevel(cmd_data)
                if cmd_id <= 0:
                    raise RuntimeError(f"Invalid cmd_id={cmd_id}")
                await asyncio.sleep(0.001)

                ack_data = SALPY_Test.Test_ackcmdC()
                while True:
                    response_id = manager.getResponse_setLogLevel(ack_data)
                    if response_id == cmd_id:
                        self.assertEqual(ack_data.identity, f"Test:{self.index}")
                        if ack_data.ack == SALPY_Test.SAL__CMD_COMPLETE:
                            return ack_data
                        elif ack_data.ack in done_ack_codes:
                            raise RuntimeError(
                                f"Remote: command failed; ack={ack_data.ack}"
                            )
                    await asyncio.sleep(0.01)

            print("Remote: wait for initial logLevel event")
            data = await asyncio.wait_for(get_logLevel(), timeout=STD_TIMEOUT)
            print(f"Remote: read logLevel.level={data.level}")
            self.assertEqual(data.level, INITIAL_LOG_LEVEL)

            for level in (10, 52, 0):
                print(f"Remote: send setLogLevel(level={level}) command")
                await asyncio.wait_for(send_setLogLevel(level), timeout=STD_TIMEOUT)

                print("Remote: wait for logLevel event")
                data = await asyncio.wait_for(get_logLevel(), timeout=STD_TIMEOUT)
                print(f"Remote: read logLevel.level={data.level}")
                self.assertEqual(data.level, level)
                print("Remote: wait for scalars telemetry")
                data = await asyncio.wait_for(get_scalars(), timeout=STD_TIMEOUT)
                print(f"Remote: read scalars.int0={data.int0}")
                self.assertEqual(data.int0, level)
                await asyncio.sleep(0.1)

            await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
        finally:
            print("Remote: done")
            if manager is not None:
                manager.salShutdown()
            if process.returncode is None:
                process.terminate()
                warnings.warn("Killed a process that was not properly terminated")


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import logging
import os
import subprocess
import unittest
import warnings

import yaml

from lsst.ts import salobj
from lsst.ts.idl.enums.Script import ScriptState

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

index_gen = salobj.index_generator()


class NonConfigurableScript(salobj.BaseScript):
    """A script that takes no configuration.

    In other words get_schema returns None.
    """

    def __init__(self, index):
        super().__init__(index=index, descr="Non-configurable script")
        self.config = None
        self.run_called = False
        self.set_metadata_called = False

    @classmethod
    def get_schema(cls):
        return None

    async def configure(self, config):
        self.config = config

    async def run(self):
        self.run_called = True

    def set_metadata(self, metadata):
        self.set_metadata_called = True


class BaseScriptTestCase(unittest.IsolatedAsyncioTestCase):
    """Test `BaseScript` using simple subclasses `TestScript` and
    `NonConfigurableScript`.
    """

    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        self.datadir = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
        self.index = next(index_gen)

    async def configure_and_check(
        self, script, log_level=0, pause_checkpoint="", stop_checkpoint="", **kwargs
    ):
        """Configure a script by calling ``do_configure`` and set group ID
        and check the result.

        Parameters
        ----------
        script : `ts.salobj.TestScript`
            A test script
        log_level : `int`, optional
            Log level as a `logging` level,
            or 0 to leave the script's log level unchanged.
        pause_checkpoint : `str`, optional
            Checkpoint(s) at which to pause, as a regular expression.
            "" to not pause at any checkpoint; "*" to pause at all checkpoints.
        stop_checkpoint : `str`, optional
            Checkpoint(s) at which to stop, as a regular expression.
            "" to not stop at any checkpoint; "*" to stop at all checkpoints.
        kwargs : `dict`
            A dict with one or more of the following keys:

            * ``wait_time`` (a float): how long to wait, in seconds
            * ``fail_run`` (bool): fail before waiting?
            * ``fail_cleanup`` (bool): fail in cleanup?

            If no values are specified then ``script.do_configure``
            is called with an empty string.

        Raises
        ------
        AssertionError
            If the script's ``config``, log level, or checkpoints
            do not match what was commanded.

        salobj.ExpectedError
            If ``kwargs`` includes other keywords than those
            documented above.
        """
        if kwargs:
            # strip to remove final trailing newline
            config = yaml.safe_dump(kwargs).strip()
        else:
            config = ""
        configure_data = script.cmd_configure.DataType()
        configure_data.config = config
        configure_data.logLevel = log_level
        configure_data.pauseCheckpoint = pause_checkpoint
        configure_data.stopCheckpoint = stop_checkpoint
        await script.do_configure(configure_data)
        self.assertEqual(script.config.wait_time, kwargs.get("wait_time", 0))
        self.assertEqual(script.config.fail_run, kwargs.get("fail_run", False))
        self.assertEqual(script.config.fail_cleanup, kwargs.get("fail_cleanup", False))
        if log_level != 0:
            self.assertEqual(script.log.getEffectiveLevel(), log_level)
        self.assertEqual(script.evt_checkpoints.data.pause, pause_checkpoint)
        self.assertEqual(script.evt_checkpoints.data.stop, stop_checkpoint)
        self.assertEqual(script.state.state, ScriptState.CONFIGURED)

        # Cannot run: groupId is not set.
        self.assertEqual(script.group_id, "")
        run_data = script.cmd_run.DataType()
        with self.assertRaises(salobj.ExpectedError):
            await script.do_run(run_data)

        # Set and check group ID.
        group_id = "arbitrary group ID"
        group_id_data = script.cmd_setGroupId.DataType()
        group_id_data.groupId = group_id
        await script.do_setGroupId(group_id_data)
        self.assertEqual(script.group_id, group_id)
        self.assertEqual(script.evt_state.data.groupId, group_id)

    def test_get_schema(self):
        schema = salobj.TestScript.get_schema()
        self.assertTrue(isinstance(schema, dict))
        for name in ("$schema", "$id", "title", "description", "type", "properties"):
            self.assertIn(name, schema)
        self.assertFalse(schema["additionalProperties"])

    def test_non_configurable_script_get_schema(self):
        schema = NonConfigurableScript.get_schema()
        self.assertIsNone(schema)

    async def test_non_configurable_script_empty_config(self):
        """Test configuring the script with no data. This should work."""
        async with NonConfigurableScript(index=self.index) as script:
            data = script.cmd_configure.DataType()
            await script.do_configure(data)
            self.assertEqual(len(script.config.__dict__), 0)
            self.assertTrue(script.set_metadata_called)
            self.assertFalse(script.run_called)

    async def test_non_configurable_script_invalid_config(self):
        async with NonConfigurableScript(index=self.index) as script:
            data = script.cmd_configure.DataType()
            data.config = "invalid: should be empty"
            with self.assertRaises(salobj.ExpectedError):
                await script.do_configure(data)
            self.assertIsNone(script.config)

    async def test_script_environ(self):
        """Test that creating a script does not modify os.environ.

        I would like to also test that a script has master priority 0
        (which is done by temporarily setting env var
        salobj.MASTER_PRIORITY_ENV_VAR to "0"),
        but that information is not available.
        """
        for master_priority in ("21", None):
            with salobj.modify_environ(
                **{salobj.MASTER_PRIORITY_ENV_VAR: master_priority}
            ):
                initial_environ = os.environ.copy()
                async with NonConfigurableScript(index=self.index):
                    self.assertEqual(os.environ, initial_environ)
                # Test again when script is closed, just to be sure;
                # closing a script should not modify the environment.
                self.assertEqual(os.environ, initial_environ)

    async def test_setCheckpoints(self):
        async with salobj.TestScript(index=self.index) as script:

            # try valid values
            data = script.cmd_setCheckpoints.DataType()
            for pause, stop in (
                ("something", ""),
                ("", "something_else"),
                (".*", "start|end"),
            ):
                data.pause = pause
                data.stop = stop
                await script.do_setCheckpoints(data)
                self.assertEqual(script.checkpoints.pause, pause)
                self.assertEqual(script.checkpoints.stop, stop)

            # try with at least one checkpoint not a valid regex;
            # do_setCheckpoints should raise and not change the checkpoints
            initial_pause = "initial_pause"
            initial_stop = "initial_stop"
            data.pause = initial_pause
            data.stop = initial_stop
            await script.do_setCheckpoints(data)
            for bad_pause, bad_stop in (("(", ""), ("", "("), ("[", "[")):
                data.pause = bad_pause
                data.stop = bad_stop
                with self.assertRaises(salobj.ExpectedError):
                    await script.do_setCheckpoints(data)
                self.assertEqual(script.checkpoints.pause, initial_pause)
                self.assertEqual(script.checkpoints.stop, initial_stop)

    async def test_set_state_and_attributes(self):
        async with salobj.TestScript(index=self.index) as script:
            # check keep_old_reason argument of set_state
            reason = "initial reason"
            additional_reason = "check append"
            script.set_state(reason=reason)
            script.set_state(reason=additional_reason, keep_old_reason=True)
            self.assertEqual(script.state.reason, reason + "; " + additional_reason)

            bad_state = 1 + max(s.value for s in ScriptState)
            with self.assertRaises(ValueError):
                script.set_state(bad_state)
            script.state.state = bad_state
            self.assertEqual(script.state_name, f"UNKNOWN({bad_state})")
            self.assertFalse(script._is_exiting)

            script.set_state(ScriptState.CONFIGURED)
            self.assertEqual(script.state_name, "CONFIGURED")

            # check assert_states
            all_states = set(ScriptState)
            for state in ScriptState:
                script.set_state(state)
                self.assertEqual(script.state_name, state.name)
                with self.assertRaises(salobj.ExpectedError):
                    script.assert_state(
                        "should fail because state not in allowed states",
                        all_states - set([state]),
                    )

                script.assert_state("should pass", [state])
                script._is_exiting = True
                with self.assertRaises(salobj.ExpectedError):
                    script.assert_state("should fail because exiting", [state])
                script._is_exiting = False

                # check that checkpoint is prohibited
                # unless state is RUNNING
                if state == ScriptState.RUNNING:
                    continue
                with self.assertRaises(RuntimeError):
                    await script.checkpoint("foo")

            self.assertFalse(script.done_task.done())

    async def test_next_supplemented_group_id(self):
        async with salobj.TestScript(index=self.index) as script:
            await self.configure_and_check(script)
            group_id = script.group_id
            for i in range(5):
                # Format some other way than an f-string,
                # in order to have a different implementation than Script
                desired_supplemented_id = "%s#%s" % (group_id, i + 1)
                supplemented_id = script.next_supplemented_group_id()
                self.assertEqual(supplemented_id, desired_supplemented_id)

            # Set a new group ID. This should reset the subgroup counter.
            new_group_id = group_id + " modified"
            group_id_data = script.cmd_setGroupId.DataType()
            group_id_data.groupId = new_group_id
            await script.do_setGroupId(group_id_data)

            for i in range(5):
                desired_supplemented_id = "%s#%s" % (new_group_id, i + 1)
                supplemented_id = script.next_supplemented_group_id()
                self.assertEqual(supplemented_id, desired_supplemented_id)

            # Clear the group ID; getting a supplemened group ID should fail
            group_id_data = script.cmd_setGroupId.DataType()
            group_id_data.groupId = ""
            await script.do_setGroupId(group_id_data)
            with self.assertRaises(RuntimeError):
                script.next_supplemented_group_id()

    async def test_pause(self):
        async with salobj.TestScript(index=self.index) as script:
            # Cannot run in UNCONFIGURED state.
            run_data = script.cmd_run.DataType()
            with self.assertRaises(salobj.ExpectedError):
                await script.do_run(run_data)

            # Test configure with data for a non-existent argument.
            configure_data = script.cmd_configure.DataType()
            configure_data.config = "no_such_arg: 1"
            with self.assertRaises(salobj.ExpectedError):
                await script.do_configure(configure_data)
            self.assertEqual(script.state.state, ScriptState.UNCONFIGURED)

            # Test configure with invalid yaml.
            configure_data = script.cmd_configure.DataType()
            configure_data.config = "a : : 2"
            with self.assertRaises(salobj.ExpectedError):
                await script.do_configure(configure_data)
            self.assertEqual(script.state.state, ScriptState.UNCONFIGURED)

            # Test configure with yaml that makes a string, not a dict.
            configure_data = script.cmd_configure.DataType()
            configure_data.config = "just_a_string"
            with self.assertRaises(salobj.ExpectedError):
                await script.do_configure(configure_data)
            self.assertEqual(script.state.state, ScriptState.UNCONFIGURED)

            # Test configure with yaml that makes a list, not a dict.
            configure_data = script.cmd_configure.DataType()
            configure_data.config = "['not', 'a', 'dict']"
            with self.assertRaises(salobj.ExpectedError):
                await script.do_configure(configure_data)
            self.assertEqual(script.state.state, ScriptState.UNCONFIGURED)

            # Now test valid configuration; specify nonexistent checkpoints
            # to test that the configure command handles checkpoints at all.
            wait_time = 0.5
            # Specify a log level that is not the default (which is INFO)
            # and is only slightly more verbose than INFO.
            log_level = logging.INFO - 1
            prelim_pause_checkpoint = "preliminary nonexistent pause checkpoint"
            prelim_stop_checkpoint = "preliminary nonexistent stop checkpoint"
            await self.configure_and_check(
                script,
                wait_time=wait_time,
                log_level=log_level,
                pause_checkpoint=prelim_pause_checkpoint,
                stop_checkpoint=prelim_stop_checkpoint,
            )

            # Set a pause checkpoint that exists.
            setCheckpoints_data = script.cmd_setCheckpoints.DataType()
            checkpoint_named_start = "start"
            checkpoint_that_does_not_exist = "nonexistent checkpoint"
            setCheckpoints_data.pause = checkpoint_named_start
            setCheckpoints_data.stop = checkpoint_that_does_not_exist
            await script.do_setCheckpoints(setCheckpoints_data)
            self.assertEqual(script.checkpoints.pause, checkpoint_named_start)
            self.assertEqual(script.checkpoints.stop, checkpoint_that_does_not_exist)

            # Run the script.
            run_data = script.cmd_run.DataType()
            run_task = asyncio.create_task(script.do_run(run_data))
            niter = 0
            while script.state.state != ScriptState.PAUSED:
                niter += 1
                await asyncio.sleep(0)
            self.assertEqual(script.state.lastCheckpoint, checkpoint_named_start)
            self.assertEqual(script.checkpoints.pause, checkpoint_named_start)
            self.assertEqual(script.checkpoints.stop, checkpoint_that_does_not_exist)
            resume_data = script.cmd_resume.DataType()
            await script.do_resume(resume_data)
            await asyncio.wait_for(run_task, 2)
            await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
            duration = (
                script.timestamps[ScriptState.ENDING]
                - script.timestamps[ScriptState.RUNNING]
            )
            desired_duration = wait_time
            print(f"test_pause duration={duration:0.2f}")
            self.assertLess(abs(duration - desired_duration), 0.2)

    async def test_stop_at_checkpoint(self):
        async with salobj.TestScript(index=self.index) as script:
            wait_time = 0.1
            await self.configure_and_check(script, wait_time=wait_time)

            # set a stop checkpoint
            setCheckpoints_data = script.cmd_setCheckpoints.DataType()
            checkpoint_named_end = "end"
            setCheckpoints_data.stop = checkpoint_named_end
            await script.do_setCheckpoints(setCheckpoints_data)
            self.assertEqual(script.checkpoints.pause, "")
            self.assertEqual(script.checkpoints.stop, checkpoint_named_end)

            run_data = script.cmd_run.DataType()
            await asyncio.wait_for(script.do_run(run_data), 2)
            await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
            self.assertEqual(script.state.lastCheckpoint, checkpoint_named_end)
            self.assertEqual(script.state.state, ScriptState.STOPPED)
            duration = (
                script.timestamps[ScriptState.STOPPING]
                - script.timestamps[ScriptState.RUNNING]
            )
            # waited and then stopped at the "end" checkpoint
            desired_duration = wait_time
            print(f"test_stop_at_checkpoint duration={duration:0.2f}")
            self.assertLess(abs(duration - desired_duration), 0.2)

    async def test_stop_while_paused(self):
        async with salobj.TestScript(index=self.index) as script:
            wait_time = 5
            await self.configure_and_check(script, wait_time=wait_time)

            # set a stop checkpoint
            setCheckpoints_data = script.cmd_setCheckpoints.DataType()
            checkpoint_named_start = "start"
            setCheckpoints_data.pause = checkpoint_named_start
            await script.do_setCheckpoints(setCheckpoints_data)
            self.assertEqual(script.checkpoints.pause, checkpoint_named_start)
            self.assertEqual(script.checkpoints.stop, "")

            run_data = script.cmd_run.DataType()
            asyncio.create_task(script.do_run(run_data))
            while script.state.lastCheckpoint != "start":
                await asyncio.sleep(0)
            self.assertEqual(script.state.state, ScriptState.PAUSED)
            stop_data = script.cmd_stop.DataType()
            await script.do_stop(stop_data)
            await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
            self.assertEqual(script.state.lastCheckpoint, checkpoint_named_start)
            self.assertEqual(script.state.state, ScriptState.STOPPED)
            duration = (
                script.timestamps[ScriptState.STOPPING]
                - script.timestamps[ScriptState.RUNNING]
            )
            # the script ran quickly because we stopped the script
            # just as soon as it paused at the "start" checkpoint
            desired_duration = 0
            print(f"test_stop_while_paused duration={duration:0.2f}")
            self.assertGreater(duration, 0.0)
            self.assertLess(abs(duration - desired_duration), 0.2)

    async def test_stop_while_running(self):
        async with salobj.TestScript(index=self.index) as script:
            wait_time = 5
            pause_time = 0.5
            await self.configure_and_check(script, wait_time=wait_time)

            checkpoint_named_start = "start"
            run_data = script.cmd_run.DataType()
            asyncio.create_task(script.do_run(run_data))
            while script.state.lastCheckpoint != checkpoint_named_start:
                await asyncio.sleep(0)
            self.assertEqual(script.state.state, ScriptState.RUNNING)
            await asyncio.sleep(pause_time)
            stop_data = script.cmd_stop.DataType()
            await script.do_stop(stop_data)
            await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
            self.assertEqual(script.state.lastCheckpoint, checkpoint_named_start)
            self.assertEqual(script.state.state, ScriptState.STOPPED)
            duration = (
                script.timestamps[ScriptState.STOPPING]
                - script.timestamps[ScriptState.RUNNING]
            )
            # we waited `pause_time` seconds after the "start" checkpoint
            desired_duration = pause_time
            print(f"test_stop_while_running duration={duration:0.2f}")
            self.assertLess(abs(duration - desired_duration), 0.2)

    async def check_fail(self, fail_run):
        """Check failure in run or cleanup.

        Parameters
        ----------
        fail_run : `bool`
            If True then fail in the script's ``run`` method,
            else fail in the script's ``cleanup`` method.
        """
        wait_time = 0.1
        async with salobj.TestScript(index=self.index) as script:
            if fail_run:
                await self.configure_and_check(script, fail_run=True)
            else:
                await self.configure_and_check(script, fail_cleanup=True)

            run_data = script.cmd_run.DataType()
            await asyncio.wait_for(script.do_run(run_data), 2)
            if fail_run:
                await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
                self.assertEqual(script.state.lastCheckpoint, "start")
                self.assertEqual(script.state.state, ScriptState.FAILED)
                end_run_state = ScriptState.FAILING
            else:
                with self.assertRaises(salobj.ExpectedError):
                    await asyncio.wait_for(script.done_task, timeout=STD_TIMEOUT)
                self.assertEqual(script.state.lastCheckpoint, "end")
                end_run_state = ScriptState.ENDING
            duration = (
                script.timestamps[end_run_state]
                - script.timestamps[ScriptState.RUNNING]
            )
            # if fail_run then failed before waiting,
            # otherwise failed after
            desired_duration = 0 if fail_run else wait_time
            print(f"test_fail duration={duration:0.3f} with fail_run={fail_run}")
            self.assertLess(abs(duration - desired_duration), 0.2)

    async def test_fail_run(self):
        await self.check_fail(fail_run=True)

    async def test_fail_cleanup(self):
        await self.check_fail(fail_run=False)

    async def test_zero_index(self):
        with self.assertRaises(ValueError):
            salobj.TestScript(index=0)

    async def test_remote(self):
        """Test a script with remotes.

        Check that the remote_names attribute of the description event
        is properly set, and that the remotes have started when the
        script has started.
        """

        class ScriptWithRemotes(salobj.TestScript):
            def __init__(self, index, remote_indices):
                super().__init__(index, descr="Script with remotes")
                remotes = []
                # use remotes that read history here, to check that
                # script.start_task waits for the start_task in each remote.
                for rind in remote_indices:
                    remotes.append(
                        salobj.Remote(domain=self.domain, name="Test", index=rind)
                    )
                self.remotes = remotes

        remote_indices = [5, 7]
        async with ScriptWithRemotes(self.index, remote_indices) as script:
            remote_name_list = [f"Test:{ind}" for ind in remote_indices]
            desired_remote_names = ",".join(sorted(remote_name_list))
            self.assertEqual(script.evt_description.data.remotes, desired_remote_names)
            for remote in script.remotes:
                self.assertTrue(remote.start_task.done())

    async def test_script_process(self):
        """Test running a script as a subprocess."""
        script_path = os.path.join(self.datadir, "script1")

        for fail in (None, "fail_run", "fail_cleanup"):
            with self.subTest(fail=fail):
                async with salobj.Domain() as domain:
                    index = next(index_gen)
                    remote = salobj.Remote(
                        domain=domain,
                        name="Script",
                        index=index,
                        evt_max_history=0,
                    )
                    await asyncio.wait_for(remote.start_task, timeout=STD_TIMEOUT)

                    def logcallback(data):
                        print(f"message={data.message}")

                    remote.evt_logMessage.callback = logcallback

                    process = await asyncio.create_subprocess_exec(
                        script_path, str(index)
                    )
                    try:
                        self.assertIsNone(process.returncode)

                        state = await remote.evt_state.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(state.state, ScriptState.UNCONFIGURED)
                        self.assertEqual(state.groupId, "")

                        logLevel_data = await remote.evt_logLevel.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(logLevel_data.level, logging.INFO)

                        wait_time = 0.1
                        config = f"wait_time: {wait_time}"
                        if fail:
                            config = config + f"\n{fail}: True"
                        await remote.cmd_configure.set_start(
                            config=config, timeout=STD_TIMEOUT
                        )
                        state = await remote.evt_state.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(state.state, ScriptState.CONFIGURED)
                        self.assertEqual(state.groupId, "")

                        metadata = await remote.evt_metadata.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(metadata.duration, wait_time)

                        group_id = "a non-blank group ID"
                        await remote.cmd_setGroupId.set_start(
                            groupId=group_id, timeout=STD_TIMEOUT
                        )
                        state = await remote.evt_state.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(state.groupId, group_id)

                        await remote.cmd_run.start(timeout=STD_TIMEOUT)

                        await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
                        if fail:
                            self.assertEqual(process.returncode, 1)
                        else:
                            self.assertEqual(process.returncode, 0)
                    finally:
                        if process.returncode is None:
                            process.terminate()
                            warnings.warn(
                                "Killed a process that was not properly terminated"
                            )

    async def test_script_schema_process(self):
        """Test running a script with --schema as a subprocess."""
        script_path = os.path.join(self.datadir, "script1")
        index = 1  # index is ignored
        process = await asyncio.create_subprocess_exec(
            script_path,
            str(index),
            "--schema",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            stdout, stderr = await asyncio.wait_for(
                process.communicate(), timeout=STD_TIMEOUT
            )
            schema = yaml.safe_load(stdout)
            self.assertEqual(schema, salobj.TestScript.get_schema())
            await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
            self.assertEqual(process.returncode, 0)
        finally:
            if process.returncode is None:
                process.terminate()
                warnings.warn("Killed a process that was not properly terminated")


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pathlib
import unittest

import yaml
import jsonschema

from lsst.ts import salobj


class ValidatorTestCase(unittest.TestCase):
    def setUp(self):
        self.validator = salobj.DefaultingValidator(schema=salobj.CONFIG_SCHEMA)

    def test_no_config_specified(self):
        data_dict = {}
        result = self.validator.validate(data_dict)
        # the defaults are hard-coded in schema/Test.yaml
        expected_result = dict(
            string0="default value for string0",
            bool0=True,
            int0=5,
            float0=3.14,
            intarr0=[-1, 1],
            multi_type=None,
        )
        self.assertEqual(data_dict, {})  # input not changed
        self.assertEqual(result, expected_result)

        result = self.validator.validate(None)
        self.assertEqual(result, expected_result)

    def test_all_fields(self):
        """Test a config with all fields set to a non-default value."""
        data_dict = dict(
            string0="an arbitrary string",
            bool0=False,
            int0=-47,
            float0=1.234,
            intarr0=[0, 2, -3, -5, 4],
            multi_type="another string",
        )
        original_data = data_dict.copy()
        result = self.validator.validate(data_dict)
        self.assertEqual(result, original_data)  # input not changed
        # all values were provided so none should be altered
        self.assertEqual(data_dict, original_data)

    def test_some_fields(self):
        """Test a config with some fields set to a non-default value."""
        default_values = dict(
            string0="default value for string0",
            bool0=True,
            int0=5,
            float0=3.14,
            intarr0=[-1, 1],
            multi_type=None,
        )
        non_default_values = dict(
            string0="an arbitrary string",
            bool0=False,
            int0=-47,
            float0=1.234,
            intarr0=[0, 2, -3, -5, 4],
            multi_type=5,
        )
        for name, value in non_default_values.items():
            expected_values = default_values.copy()
            expected_values[name] = value
            one_item_data_dict = {name: value}
        result = self.validator.validate(one_item_data_dict)
        self.assertEqual(result, expected_values)

    def test_invalid_data(self):
        good_data = dict(
            string0="an arbitrary string",
            bool0=False,
            int0=-47,
            float0=1.234,
            intarr0=[0, 2, -3, -5, 4],
            multi_type=5,
        )
        bad_data = dict(
            string0=45, bool0=35, int0=1.234, float0="hello", intarr0=45, multi_type=3.5
        )
        # set one field at a time to bad data
        for field in good_data:
            data = good_data.copy()
            data[field] = bad_data[field]
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                self.validator.validate(data)

        # all fields are invalid
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            self.validator.validate(bad_data)

        # extra field
        bad_data = good_data.copy()
        bad_data["unwanted"] = 0
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            self.validator.validate(bad_data)

        # invalid types
        for bad_type_data in (True, False, 1, 1.34, "hello", (1, 2), ["a", "b"]):
            with self.subTest(bad_type_data=bad_type_data):
                with self.assertRaises(jsonschema.exceptions.ValidationError):
                    self.validator.validate(bad_type_data)


class DefaultingTestCase(unittest.TestCase):
    def setUp(self):
        self.schemadir = pathlib.Path(__file__).resolve().parents[0] / "data"

    def test_contained_object(self):
        schemapath = self.schemadir / "contained_object_schema.yaml"
        with open(schemapath, "r") as f:
            rawschema = f.read()
        self.schema = yaml.safe_load(rawschema)
        self.validator = salobj.DefaultingValidator(schema=self.schema)
        default_values = self.validator.validate({})
        print(f"default_values={default_values}")
        self.assertEqual(default_values["number1"], 1)
        self.assertEqual(default_values["subobject"]["subnumber1"], 2)
        import types

        print(types.SimpleNamespace(**default_values))


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["stream_as_generator", "CscCommander"]

import argparse
import asyncio
import collections
import functools
import shlex
import sys
import warnings

from . import domain
from . import remote
from . import sal_enums
from . import csc_utils

# A dict of valid values for bool command arguments.
# The argument should be converted to lowercase before using.
BOOL_DICT = {
    "0": False,
    "f": False,
    "false": False,
    "1": True,
    "t": True,
    "true": True,
}


async def stream_as_generator(stream, encoding="utf-8"):
    """Await lines of text from stdin or another input stream.

    Example usage:

        async for line in stream_as_generator(stream=sys.stdin):
            print(f"read {repr(line)}")

    Parameters
    ----------
    stream : ``stream``
        Stream to read, e.g. `sys.stdin`.
    encoding : `str` or `None`
        Encoding. If provided then decode the line,
        else return the line as raw bytes.

    Returns
    -------
    line : `str`
        A line of data, optionally decoded.

    Notes
    -----
    Thanks to
    http://blog.mathieu-leplatre.info/some-python-3-asyncio-snippets.html
    """
    loop = asyncio.get_running_loop()
    reader = asyncio.StreamReader(loop=loop)
    reader_protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: reader_protocol, sys.stdin)
    while True:
        line = await reader.readline()
        if not line:  # EOF.
            break
        if encoding is not None:
            line = line.decode(encoding)
        yield line


def round_any(value, digits=2):
    """Round any value to the specified number of digits.

    This is a no-op for int and str values.

    Notes
    -----
    This is designed for use with DDS samples. As such, it expects
    array values to be of type `list`.
    """
    if isinstance(value, float):
        return round(value, digits)
    elif value and isinstance(value, list) and isinstance(value[0], float):
        return [round(item, digits) for item in value]
    return value


class CscCommander:
    """Command a CSC from the command line.

    Parameters
    ----------
    name : `str`
        SAL component name of CSC.
    index : `int`, optional
        SAL index of CSC.
    enable : `bool`, optional
        Enable the CSC (when the commander starts up)?
        Note: `amain` always supplies this argument.
    exclude : `List` [`str`] or `None`, optional
        Names of telemetry or event topics to not print.
        If `None` or empty then no topics are excluded.
    exclude_commands : `List` [`str`], optional
        Names of commands to ignore. Typically this should include:

        * "abort" and "setValue", two generic commands that few CSCs implement.
        * "enterControl" unless your CSC is "externally commandable",
          meaning it can run in the OFFLINE state.

        The default is to exclude nothing, for backwards compatibility.
    fields_to_ignore : `List` [`str`], optional
        SAL topic fields names to ignore when specifying command parameters,
        and when printing events and telemetry.
    telemetry_fields_to_not_compare : `List` [`str`], optional
        Telemetry field names to ignore when checking if a telemetry topic has
        changed enough to print the latest sample. These fields are printed.

    Attributes
    ----------
    domain : `Domain`
        DDS domain.
    remote : `Remote`
        Remote for the CSC being commanded.
    help_dict : `dict`
        Dict of command_name: documentation.
        You should add one entry for every do_command method you define.
        Each documentation string should start with a list of argument names
        (the command name will be prepended for you)
        optionally followed by # and a brief description.
        The documentation string should a single line, if practical.


    Notes
    -----
    Warning: use with caution. Running a commander may interfere with
    telescope operations! If you provide a command-line script in ``bin/``
    that runs a commander, consider picking a name that is obscure
    and not easily confused with the script that runs the CSC.

    Make a subclass of `CscCommander` if you want to add extra commands
    or provide special handling for some commands, events or telemetry.
    Subclasses may provide overrides as follows:

    * To hide unwanted commands: delete them from ``self.command_dict``
      in your constructor. Most CSCs should hide the "abort", "enterControl"
      and "setValue" commands, as follows::

        for command_to_ignore in ("abort", "enterControl", "setValue"):
            del self.command_dict[command_to_ignore]

    * To override handling of a standard command (one defined in the XML):
      define a ``do_<command_name>`` method.
      The method receives one argument: a list of str arguments.
      You should provide a custom handler for, or hide, any command with
      array parameters, because the command parser only accepts scalars.

    * To add an additional command (one not defined in the XML):

        * Define a  ``do_<command_name>`` to handle the command.
          The method receives one argument: a list of str arguments.

        * Add an entry to ``help_dict``. The key is the command name
          and the value is a brief (preferably only one line) help string
          that lists the arguments first, and a brief description after.
          Here is an example::

            self.help_dict["sine"] = "start_position amplitude "
            "# track one cycle of a sine wave",

    * To override handling of an event or telemetry topic: define method
      ``evt_<event_name>_callback`` or ``tel_<event_name>_callback``,
      respectively.  It receives one argument: the DDS sample.
      This can be especially useful if the default behavior is too chatty
      for one or more telemetry topics.

    I have not found a way to write a unit test for this class.
    I tried running a commander in a subprocess but could not figure out
    how to send multiple commands (the ``suprocess.communicate``
    method only allows sending one item of data).
    Instead I suggest manually running it to control the Test CSC.

    Examples
    --------

    You can do a lot with just the base class. This command-line script
    will control the Test CSC pretty well::

        import asyncio
        from lsst.ts import salobj
        asyncio.run(salobj.CscCommander.amain(name="Test", index=True))

    However, this will claim to support some generic commands that the
    Test CSC does not support ( "abort", "enterControl", and "setValue")
    and it mishandles the "setArrays" command.
    See `TestCscCommander` for a version that fixes these deficiencies.
    `TestCscCommander` is run as follows::

        import asyncio
        from lsst.ts import salobj
        asyncio.run(salobj.TestCscCommander.amain(index=True))

    For an example with extra commands and special telemetry handling,
    see ``RotatorCommander`` in the ts_rotator package.

    **Unit Testing**

    To unit test a commander:

    * Set ``commander.testing = True``.
      This appends all output messages to output queue
      ``commander.output_queue``, a `collections.deque`.
      The messages are also printed to ``stdout``, as usual.
    * Call ``commander.run_command`` to run a command.
    * Check the output using the output queue.
    * Clear the output queue with ``commander.output_queue.clear()``
      whenever you like, e.g. before running a command whose output
      you want to check.
    """

    def __init__(
        self,
        name,
        index=0,
        enable=False,
        exclude=None,
        exclude_commands=(),
        fields_to_ignore=("ignored", "value", "priority"),
        telemetry_fields_to_not_compare=("timestamp",),
    ):
        self.domain = domain.Domain()
        self.remote = remote.Remote(
            domain=self.domain, name=name, index=index, exclude=exclude
        )
        self.fields_to_ignore = frozenset(fields_to_ignore)
        self.telemetry_fields_to_not_compare = frozenset(
            telemetry_fields_to_not_compare
        )
        self.tasks = set()
        self.help_dict = dict()
        self.enable = enable
        self.testing = False
        self.output_queue = collections.deque()

        for name in self.remote.salinfo.event_names:
            if name == "heartbeat":
                continue
            topic_attr_name = f"evt_{name}"
            topic = getattr(self.remote, topic_attr_name)
            callback = getattr(self, f"{topic_attr_name}_callback", None)
            if callback is None:
                callback = functools.partial(self.event_callback, name=name)
            setattr(topic, "callback", callback)

        for name in self.remote.salinfo.telemetry_names:
            topic_attr_name = f"tel_{name}"
            setattr(self, f"previous_{topic_attr_name}", None)
            topic = getattr(self.remote, topic_attr_name)
            callback = getattr(self, f"{topic_attr_name}_callback", None)
            if callback is None:
                callback = functools.partial(self.telemetry_callback, name=name)
            setattr(topic, "callback", callback)

        # Dict of command name: RemoteCommand topic:
        self.command_dict = {
            name: getattr(self.remote, f"cmd_{name}")
            for name in self.remote.salinfo.command_names
            if name not in frozenset(exclude_commands)
        }

    def print_help(self):
        """Print help."""
        command_help = "\n".join(self.get_commands_help())
        self.output(
            f"""Send commands to the {self.remote.salinfo.name} CSC and print events and telemetry.

CSC Commands:
{command_help}
exit  # exit this commander (leaving the CSC running)
help  # print this help
"""
        )

    async def close(self):
        """Close the commander, prior to quitting."""
        while self.tasks:
            task = self.tasks.pop()
            task.cancel()
        await self.remote.close()
        await self.domain.close()

    def output(self, str):
        """Print a string to output, appending a final newline.

        Please call this instead of print to support unit tests.

        If ``self.testing`` is True then append ``str`` to
        ``self.output_queue`` instead of printing it.
        Use this mode for unit testing a CSC commander.
        """
        if self.testing:
            self.output_queue.append(str)
        print(str)

    def format_item(self, key, value):
        """Format one event or telemetry field for printing."""
        if isinstance(value, float):
            return f"{key}={value:0.4f}"
        return f"{key}={value}"

    def format_data(self, data):
        """Format the public fields of an event or telemetry sample,
        for printing.
        """
        return self.format_dict(self.get_public_data(data))

    def format_dict(self, data):
        """Format a dict for printing.

        Unlike format_data, this requires a dict and formats *all* fields.
        """
        return ", ".join(self.format_item(key, value) for key, value in data.items())

    def field_is_public(self, name):
        """Return True if the specified field name is public,
        False otherwise.
        """
        if name.startswith("private_"):
            return False
        if name == f"{self.remote.salinfo.name}ID":
            return False
        if name in self.fields_to_ignore:
            return False
        return True

    def get_public_data(self, data):
        """Get a dict of field_name: value for public fields of a DDS sample.

        Parameters
        ----------
        data : ``dds_sample``
            DDS sample.
        """
        return dict(
            (key, value)
            for key, value in data.get_vars().items()
            if self.field_is_public(key)
        )

    def get_rounded_public_data(self, data, digits=2):
        """Get a dict of field_name: value for public fields of a DDS sample
        with float values rounded.
        """
        return {
            key: round_any(value, digits=digits)
            for key, value in data.get_vars().items()
            if self.field_is_public(key)
        }

    def get_rounded_public_fields(self, data, digits=2):
        """Deprecated version of get_rounded_public_data."""
        warnings.warn("Use get_rounded_public_data instead", DeprecationWarning)
        return self.get_rounded_public_data(data=data, digits=digits)

    def event_callback(self, data, name):
        """Generic callback for events.

        You may provide evt_<event_name> methods to override printing
        of specific events.
        """
        self.output(f"{data.private_sndStamp:0.3f}: {name}: {self.format_data(data)}")

    def evt_summaryState_callback(self, data):
        try:
            state = sal_enums.State(data.summaryState)
        except Exception:
            state = f"{data.summaryState} (not a known state!)"
        self.output(
            f"{data.private_sndStamp:0.3f}: summaryState: summaryState={state!r}"
        )

    def telemetry_callback(self, data, name, digits=2):
        """Generic callback for telemetry.

        You may provide tel_<telemetry_name> methods to override printing
        of specific telemetry topics.
        """
        prev_value_name = f"previous_tel_{name}"
        public_dict = self.get_rounded_public_data(data, digits=digits)
        trimmed_dict = {
            name: value
            for name, value in public_dict.items()
            if name not in self.telemetry_fields_to_not_compare
        }
        if trimmed_dict != getattr(self, prev_value_name):
            setattr(self, prev_value_name, trimmed_dict)
            formatted_data = self.format_dict(public_dict)
            self.output(f"{data.private_sndStamp:0.3f}: {name}: {formatted_data}")

    def check_arguments(self, args, *names):
        """Check that the required arguments are provided,
        and return them as a keyword argument dict with cast values.

        Parameters
        ----------
        args : `List` [`str`]
            Command arguments, as strings.
        *names : `List` [`str` or `tuple`]
            Argument name and optional cast function. Each element is either:

            * An argument name, in which case the argument is cast to a float
            * A tuple of (name, cast function), in which case the argument
                is cast using the cast function.
        """
        required_num_args = len(names)
        if len(args) != required_num_args:
            if required_num_args == 0:
                raise RuntimeError("no arguments allowed")
            else:
                raise RuntimeError(
                    f"{required_num_args} arguments required:  "
                    f"{names}; {len(args)} provided."
                )

        def cast(name, arg):
            if isinstance(name, tuple):
                if len(name) != 2:
                    raise RuntimeError(
                        "Cannot parse {name} as (name, casting function)"
                    )
                arg_name, cast_func = name
                return (arg_name, cast_func(arg))
            else:
                return (name, float(arg))

        return dict(cast(name, arg) for name, arg in zip(names, args))

    async def do_start(self, args):
        """Allow the start command to have no arguments."""
        assert len(args) in (0, 1)
        if args:
            settingsToApply = args[0]
        else:
            settingsToApply = ""
        await self.remote.cmd_start.set_start(
            settingsToApply=settingsToApply,
        )

    def get_commands_help(self):
        """Get help for each command, as a list of strings.

        End with "Other Commands:" and any commands
        in help_dict that are not in command_dict.
        """
        help_strings = []
        for command_name in sorted(self.command_dict.keys()):
            field_names_str = self.help_dict.get(command_name, None)
            if field_names_str is None:
                command_topic = self.command_dict[command_name]
                sample = command_topic.DataType()
                public_dict = self.get_public_data(sample)
                field_names_str = " ".join(public_dict.keys())
            help_strings.append(f"{command_name} {field_names_str}")

        other_command_names = sorted(
            command_name
            for command_name in self.help_dict
            if command_name not in self.command_dict
        )
        help_strings += ["", "Other Commands:"]
        help_strings += [
            f"{command_name} {self.help_dict[command_name]}"
            for command_name in other_command_names
        ]
        return help_strings

    async def run_command_topic(self, command_name, args):
        """Run a command that has an associated salobj RemoteCommand topic.

        Parameters
        ----------
        command_name : `str`
            Command name, e.g. Enable
        args : `List` [`str`]
            String arguments for the command.
            There must be exactly one argument per public fields.

        Notes
        -----
        This method works for command topics that take scalar arguments.
        To support command topics with more exotic arguments you must
        provide a do_<command> method that parses the arguments
        and add an entry to self.help_dict.
        """
        command = self.command_dict[command_name]
        sample = command.DataType()
        kwargs = self.get_public_data(sample)
        if len(kwargs) != len(args):
            raise ValueError(
                f"Command {command_name} requires "
                f"{len(kwargs)} arguments; got {len(args)}"
            )
        for (name, default_value), str_value in zip(kwargs.items(), args):
            try:
                if type(default_value) is bool:
                    kwargs[name] = BOOL_DICT[str_value.lower()]
                else:
                    kwargs[name] = type(default_value)(str_value)
            except Exception:
                raise ValueError(
                    f"Could not parse value {str_value!r} "
                    f"as type {type(default_value).__name__} "
                    f"for argument {name}"
                )
        await command.set_start(**kwargs)

    async def run_command(self, cmd):
        """Run the specified command string and wait for it to finish.

        Parameters
        ----------
        cmd : `str`
            Command string (command name and arguments).
            Note: does not handle the "exit" command.
        """
        tokens = shlex.split(cmd)
        command_name = tokens[0]
        args = tokens[1:]
        command_method = getattr(self, f"do_{command_name}", None)
        coro = None
        if command_name == "help":
            self.print_help()
        elif command_method is not None:
            coro = command_method(args)
        elif command_name in self.command_dict:
            coro = self.run_command_topic(command_name, args)
        else:
            self.output(f"Unrecognized command: {command_name}")

        if coro is None:
            return
        await coro

    async def _run_command_and_output(self, cmd):
        """Execute a command and wait for it to finish. Output the result.

        A wrapper around `run_command` that adds a task to self.tasks
        and catches and outputs exceptions. For use by the interactive
        command loop.

        Parameters
        ----------
        cmd : `str`
            Command string (command name and arguments).
            Note: does not handle the "exit" command.
        """
        task = asyncio.create_task(self.run_command(cmd))
        self.tasks.add(task)
        try:
            await task
            command_name = cmd.split()[0]
            self.output(f"Finished command {command_name}")
        except Exception as e:
            self.output(str(e))
        finally:
            self.tasks.remove(task)

    async def start(self):
        """Start asynchonous processes."""
        self.output(f"Waiting for {self.remote.salinfo.name_index} to start.")
        await self.remote.start_task
        if self.enable:
            self.output(f"Enabling {self.remote.salinfo.name_index}")
            # Temporarily remove the ``evt_summaryState`` callback
            # so the `set_summary_state` function can read the topic.
            summary_state_callback = self.remote.evt_summaryState.callback
            try:
                self.remote.evt_summaryState.callback = None
                await csc_utils.set_summary_state(self.remote, sal_enums.State.ENABLED)
            finally:
                summary_state = self.remote.evt_summaryState.get()
                if summary_state is not None:
                    summary_state_callback(summary_state)
                self.remote.evt_summaryState.callback = summary_state_callback

    @classmethod
    async def amain(cls, *, index, **kwargs):
        """Construct the commander and run it.

        Parse the command line to construct the commander,
        then parse and execute commands until the ``exit`` is seen.

        Parameters
        ----------
        index : `int`, `enum.IntEnum`, `True`, or `None`
            If the CSC is indexed, do one of the following:

            * Specify `True` to make ``index`` a required
              command-line argument that accepts any nonzero index.
            * Specify an `enum.IntEnum` *class* to make ``index`` a required
              command-line argument that only accepts the enum values.
            * Specify a non-zero integer to use that index.
              This is rare; if the CSC is indexed then the user
              should usually be allowed to specify the index.

            If the CSC is not indexed, specify `None` or 0.
        **kwargs : `dict`, optional
            Additional keyword arguments for your CSC's constructor.
        """
        self = cls.make_from_cmd_line(index=index, **kwargs)
        try:
            await self.start()

            self.print_help()
            async for line in stream_as_generator(sys.stdin):
                # Purge done tasks
                self.tasks = {task for task in self.tasks if not task.done()}

                # Execute the new command
                line = line.strip()
                if not line:
                    continue
                if line == "exit":
                    break
                task = asyncio.create_task(self._run_command_and_output(line))
                self.tasks.add(task)
        finally:
            await self.close()

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser created by `make_from_cmd_line`.

        Parameters
        ----------
        parser : `argparse.ArgumentParser`
            The argument parser.

        Notes
        -----
        If you override this method then you should almost certainly override
        `add_kwargs_from_args` as well.
        """
        pass

    @classmethod
    def add_kwargs_from_args(cls, args, kwargs):
        """Add constructor keyword arguments based on parsed arguments.

        Parameters
        ----------
        args : `argparse.namespace`
            Parsed command.
        kwargs : `dict`
            Keyword argument dict for the constructor.
            Update this based on ``args``.
            The index argument will already be present if relevant.

        Notes
        -----
        If you override this method then you should almost certainly override
        `add_arguments` as well.
        """
        pass

    @classmethod
    def make_from_cmd_line(cls, index, **kwargs):
        """Construct a SAL-related class from command line arguments.

        Parameters
        ----------
        index : `int`, `enum.IntEnum`, `True`, or `None`
            If the CSC is indexed, do one of the following:

            * Specify `True` to make ``index`` a required
              command-line argument that accepts any nonzero index.
            * Specify an `enum.IntEnum` *class* to make ``index`` a required
              command-line argument that only accepts the enum values.
            * Specify a non-zero integer to use that index.
              This is rare; if the CSC is indexed then the user
              should usually be allowed to specify the index.

            If the CSC is not indexed, specify `None` or 0.
        **kwargs : `dict`, optional
            Additional keyword arguments for your class's constructor.

        Returns
        -------
        instance : ``cls``
            The constructed instance.

        Notes
        -----
        To add additional command-line arguments, override `add_arguments`
        and `add_kwargs_from_args`.
        """
        parser = argparse.ArgumentParser(f"Run {cls.__name__}")
        if index is True:
            parser.add_argument("index", type=int, help="Script SAL Component index.")
        parser.add_argument(
            "-e", "--enable", action="store_true", help="Enable the CSC?"
        )
        cls.add_arguments(parser)

        args = parser.parse_args()
        if index is True:
            kwargs["index"] = args.index
        elif not index:
            pass
        else:
            kwargs["index"] = int(index)
        kwargs["enable"] = args.enable
        cls.add_kwargs_from_args(args=args, kwargs=kwargs)

        return cls(**kwargs)

    async def __aenter__(self):
        await self.remote.start_task
        return self

    async def __aexit__(self, type, value, traceback):
        await self.close()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import contextlib
import pathlib
import time
import unittest

import ddsutil
import astropy.units as u

try:
    from lsst import verify
except ImportError:
    verify = None

from lsst.ts import salobj

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

index_gen = salobj.index_generator()


class SpeedTestCase(unittest.IsolatedAsyncioTestCase):
    """Measure the speed of salobj.

    These are not actually unit tests, but running them as tests
    makes sure they are run automatically and avoids bit rot.
    """

    @classmethod
    def setUpClass(cls):
        if verify is None:
            return
        metrics = (
            verify.Metric(
                name="salobj.CreateClasses",
                description="The number of topic classes that dds can create "
                "(from an IDL file), per second. "
                "This is measured by creating one of each Test topic. "
                "Creating classes for topics is done once, at startup time.",
                unit=u.ct / u.second,
            ),
            verify.Metric(
                name="salobj.IssueCommands",
                description="The rate at which salobj can issue commands and await a reply. "
                "This test uses a command that has only one parameter, "
                "and the controller does nothing in reponse to the command.",
                unit=u.ct / u.second,
            ),
            verify.Metric(
                name="salobj.ReadTest_arrays",
                description="The rate at which salobj can read Test_arrays samples. "
                "This represents a fairly large sample.",
                unit=u.ct / u.second,
            ),
            verify.Metric(
                name="salobj.ReadTest_logLevel",
                description="The rate at which salobj can read Test_logevent_logLevel samples. "
                "This represents a small sample.",
                unit=u.ct / u.second,
            ),
            verify.Metric(
                name="salobj.WriteTest_arrays",
                description="The rate at which salobj can write Test_arrays samples. "
                "This represents a fairly large sample.",
                unit=u.ct / u.second,
            ),
            verify.Metric(
                name="salobj.WriteTest_logLevel",
                description="The rate at which salobj can write Test_logevent_logLevel samples. "
                "This represents a small sample.",
                unit=u.ct / u.second,
            ),
        )
        cls.verify_job = verify.Job(metrics=metrics)

    @classmethod
    def tearDownClass(cls):
        if verify is None:
            return
        measurements_dir = pathlib.Path(__file__).resolve().parent / "measurements"
        cls.verify_job.write(measurements_dir / "speed.json")

    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        self.datadir = pathlib.Path(__file__).resolve().parent / "data"
        self.index = next(index_gen)

    def insert_measurement(self, measurement):
        measurement.metric = self.verify_job.metrics[measurement.metric_name]
        self.verify_job.measurements.insert(measurement)

    @contextlib.asynccontextmanager
    async def make_remote_and_topic_writer(self):
        """Make a remote and launch a topic writer in a subprocess.

        Return the remote.
        """
        script_path = self.datadir / "topic_writer.py"
        process = await asyncio.create_subprocess_exec(
            str(script_path), str(self.index)
        )
        try:
            async with salobj.Domain() as domain, salobj.Remote(
                domain=domain, name="Test", index=self.index
            ) as remote:
                yield remote
                await salobj.set_summary_state(
                    remote=remote, state=salobj.State.OFFLINE, timeout=STD_TIMEOUT
                )
                await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
        finally:
            if process.returncode is None:
                print("Warning: killing the topic writer")
                process.kill()

    async def test_class_creation_speed(self):
        """Test the speed of creating topic classes on the fly."""
        async with salobj.Domain() as domain:
            salinfo = salobj.SalInfo(domain, "Test", index=self.index)
            topic_names = (
                ["logevent_" + name for name in salinfo.event_names]
                + ["command_" + name for name in salinfo.command_names]
                + list(salinfo.telemetry_names)
            )
            t0 = time.monotonic()
            for topic_name in topic_names:
                revname = salinfo.revnames.get(topic_name)
                ddsutil.get_dds_classes_from_idl(salinfo.metadata.idl_path, revname)
            dt = time.monotonic() - t0
            ntopics = len(topic_names)
            creation_speed = ntopics / dt
            print(
                f"Created {creation_speed:0.1f} topic classes/sec ({ntopics} topic classes)"
            )
            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.CreateClasses", creation_speed * u.ct / u.second
                    )
                )

    async def test_command_speed(self):
        async with self.make_remote_and_topic_writer() as remote:
            await remote.evt_summaryState.next(flush=False, timeout=60)
            t0 = time.monotonic()
            num_commands = 1000
            for i in range(num_commands):
                await remote.cmd_fault.start(timeout=STD_TIMEOUT)
            dt = time.monotonic() - t0
            command_speed = num_commands / dt
            print(
                f"Issued {command_speed:0.0f} fault commands/second ({num_commands} commands)"
            )

            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.IssueCommands", command_speed * u.ct / u.second
                    )
                )

    async def test_read_speed(self):
        async with self.make_remote_and_topic_writer() as remote:
            await salobj.set_summary_state(
                remote=remote,
                state=salobj.State.ENABLED,
                settingsToApply="arrays",
                timeout=STD_TIMEOUT,
            )

            num_samples = 1000

            # Wait for the first sample so we know the writer is running.
            data = await remote.tel_arrays.next(flush=False, timeout=STD_TIMEOUT)
            t0 = time.monotonic()
            for i in range(num_samples):
                data = await remote.tel_arrays.next(flush=False, timeout=STD_TIMEOUT)
            dt = time.monotonic() - t0
            arrays_read_speed = num_samples / dt
            nlost = data.int0[0] - num_samples
            print(
                f"Read {arrays_read_speed:0.0f} arrays samples/second "
                f"({num_samples} samples); "
                f"lost {nlost} samples"
            )

            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.ReadTest_arrays",
                        arrays_read_speed * u.ct / u.second,
                    )
                )

            await salobj.set_summary_state(
                remote=remote, state=salobj.State.STANDBY, timeout=STD_TIMEOUT
            )
            await salobj.set_summary_state(
                remote=remote,
                state=salobj.State.ENABLED,
                settingsToApply="logLevel",
                timeout=STD_TIMEOUT,
            )

            # Wait for the first logLevel sample. This is automatically
            # output by the Controller and probably has level=20
            # (unless data has been lost).
            await remote.evt_logLevel.next(flush=False, timeout=STD_TIMEOUT)
            # Wait for the next logLevel sample, which is the first one
            # output by the write loop (unless data has been lost).
            await remote.evt_logLevel.next(flush=False, timeout=STD_TIMEOUT)
            t0 = time.monotonic()
            for i in range(num_samples):
                data = await remote.evt_logLevel.next(flush=False, timeout=STD_TIMEOUT)
            dt = time.monotonic() - t0
            log_level_read_speed = num_samples / dt
            nlost = data.level - num_samples
            print(
                f"Read {log_level_read_speed:0.0f} logLevel samples/second "
                f"({num_samples} samples); "
                f"lost {nlost} samples"
            )

            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.ReadTest_logLevel",
                        log_level_read_speed * u.ct / u.second,
                    )
                )

    async def test_write_speed(self):
        async with salobj.Controller(
            name="Test", index=self.index, do_callbacks=False
        ) as controller:

            num_samples = 1000

            t0 = time.monotonic()
            for i in range(num_samples):
                controller.tel_arrays.put()
                # Free the event loop, since writers usually should.
                await asyncio.sleep(0)
            dt = time.monotonic() - t0
            arrays_write_speed = num_samples / dt
            print(
                f"Wrote {arrays_write_speed:0.0f} arrays samples/second ({num_samples} samples)"
            )

            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.WriteTest_arrays", arrays_write_speed * u.ct / u.second
                    )
                )

            t0 = time.monotonic()
            for i in range(num_samples):
                controller.evt_logLevel.put()
                await asyncio.sleep(0)
            dt = time.monotonic() - t0
            log_level_write_speed = num_samples / dt
            print(
                f"Wrote {log_level_write_speed:0.0f} logLevel samples/second ({num_samples} samples)"
            )

            if verify is not None:
                self.insert_measurement(
                    verify.Measurement(
                        "salobj.WriteTest_logLevel",
                        log_level_write_speed * u.ct / u.second,
                    )
                )


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import logging
import unittest

from lsst.ts import salobj


class QueueCapacityCheckerTestCase(unittest.TestCase):
    def setUp(self):
        self.log = logging.getLogger()

    def test_constructor(self):
        for queue_len, expected_warn_thresholds in (
            (10, (5, 9, 10)),
            (19, (5, 17, 19)),
            (20, (5, 10, 18, 20)),
            (100, (10, 50, 90, 100)),
        ):
            with self.subTest(queue_len=queue_len):
                qlc = salobj.topics.read_topic.QueueCapacityChecker(
                    descr="test", log=self.log, queue_len=queue_len
                )
                self.assertEqual(qlc.warn_thresholds, expected_warn_thresholds)

        for invalid_len in (-1, 0, 1, 9):
            with self.subTest(invalid_len=invalid_len):
                with self.assertRaises(ValueError):
                    salobj.topics.read_topic.QueueCapacityChecker(
                        descr="test", log=self.log, queue_len=invalid_len
                    )

    def test_check_nitems(self):
        """Test the check_nitems method, going from all possible
        initial states to all possible final states.
        """
        for queue_len in (10, 33, 100):
            nthresh = 3 if queue_len < 20 else 4
            # ``start_index`` and ``end_index`` are indices into
            # warning_thresholds that specify the initial and final warning
            # thresholds (and associated reset thresholds).
            # ``use_min`` specifies whether to call ``check_nitems``
            # with the minimum (True) or maximum (False) value of ``nitems``
            # that should cause the desired transition.
            for start_index, end_index, use_min in itertools.product(
                range(nthresh + 1), range(nthresh + 1), (False, True)
            ):
                with self.subTest(
                    queue_len=queue_len, start_index=start_index, end_index=end_index
                ):
                    if end_index == start_index:
                        # This case is handled by check_nochange_values
                        continue
                    qlc = self.make_checker(queue_len, start_index)
                    self.assertEqual(len(qlc.warn_thresholds), nthresh)
                    self.check_nochange_values(qlc)

                    if end_index > start_index:
                        continue
                        # Warn of an increase
                        if use_min:
                            nitems = qlc.warn_thresholds[end_index - 1]
                        elif end_index == nthresh:
                            # There is only one valid value of nitems
                            # and it is covered by use_min True
                            continue
                        else:
                            nitems = qlc.warn_thresholds[end_index] - 1
                        expected_log_level = (
                            logging.WARNING if end_index < nthresh else logging.ERROR
                        )
                        with self.assertLogs(logger=qlc.log, level=expected_log_level):
                            did_log = qlc.check_nitems(nitems)
                        self.assertTrue(did_log)
                        if end_index == nthresh:
                            expected_warn_threshold = None
                        else:
                            expected_warn_threshold = qlc.warn_thresholds[end_index]
                        self.assertEqual(qlc.warn_threshold, expected_warn_threshold)
                        expected_reset_threshold = (
                            qlc.warn_thresholds[end_index - 1] // 2
                        )
                        self.assertEqual(qlc.reset_threshold, expected_reset_threshold)
                    else:
                        # Reset to a lower warning level
                        if use_min:
                            if end_index > 0:
                                nitems = (qlc.warn_thresholds[end_index - 1] // 2) + 1
                            else:
                                nitems = 0
                        else:
                            nitems = qlc.warn_thresholds[end_index] // 2
                        did_log = qlc.check_nitems(nitems)
                        self.assertFalse(did_log)
                        expected_warn_threshold = qlc.warn_thresholds[end_index]
                        self.assertEqual(qlc.warn_threshold, expected_warn_threshold)
                        if end_index > 0:
                            expected_reset_threshold = (
                                qlc.warn_thresholds[end_index - 1] // 2
                            )
                        else:
                            expected_reset_threshold = None
                        self.assertEqual(qlc.reset_threshold, expected_reset_threshold)

    def make_checker(self, queue_len, warn_index):
        """Make a QueueCapacityChecker with specified warn_threshold.

        Parameters
        ----------
        queue_len : `int`
            Queue length. Must be >= 10.
        warn_index : `int`
            Initial warn_threshold is ``warn_thresholds[warn_index]``,
            or `None` if ``warn_index == len(warn_thresholds)``.
            then the initial warn_threshold is `None`


        """
        qlc = salobj.topics.read_topic.QueueCapacityChecker(
            descr="test", log=self.log, queue_len=queue_len
        )
        self.assertIsNone(qlc.reset_threshold)
        self.assertEqual(qlc.warn_threshold, qlc.warn_thresholds[0])
        nthresh = len(qlc.warn_thresholds)
        if warn_index > 0:
            nitems = qlc.warn_thresholds[warn_index - 1]
            expected_log_level = (
                logging.WARNING if warn_index < nthresh else logging.ERROR
            )
            with self.assertLogs(logger=qlc.log, level=expected_log_level):
                did_log = qlc.check_nitems(nitems)
            self.assertTrue(did_log)

            if warn_index == nthresh:
                expected_warn_threshold = None
            else:
                expected_warn_threshold = qlc.warn_thresholds[warn_index]
            self.assertEqual(qlc.warn_threshold, expected_warn_threshold)

            expected_reset_threshold = qlc.warn_thresholds[warn_index - 1] // 2
            self.assertEqual(qlc.reset_threshold, expected_reset_threshold)
        return qlc

    def check_nochange_values(self, queue_len_checker):
        """Check that calling `check_nitems` does nothing for the full range
        of values that should not change anything.
        """
        min_len = (
            0
            if queue_len_checker.reset_threshold is None
            else queue_len_checker.reset_threshold + 1
        )
        max_len = (
            queue_len_checker.queue_len
            if queue_len_checker.warn_threshold is None
            else queue_len_checker.warn_threshold
        )
        for n in range(min_len, max_len):
            self.assertFalse(queue_len_checker.check_nitems(n))


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["BaseScript"]

import abc
import argparse
import asyncio
import os
import re
import sys
import types

import yaml

from . import base
from . import controller
from . import validator
from lsst.ts.idl.enums.Script import (
    MetadataCoordSys,
    MetadataRotSys,
    MetadataDome,
    ScriptState,
)


HEARTBEAT_INTERVAL = 5  # seconds


def _make_remote_name(remote):
    """Make a remote name from a remote, for output as script metadata.

    Parameters
    ----------
    remote : `salobj.Remote`
        Remote
    """
    name = remote.salinfo.name
    index = remote.salinfo.index
    if index is not None:
        name = name + ":" + str(index)
    return name


class BaseScript(controller.Controller, abc.ABC):
    """Abstract base class for :ref:`lsst.ts.salobj_sal_scripts`.

    Parameters
    ----------
    index : `int`
        Index of SAL Script component. This must be non-zero
        and should be unique among all loaded SAL scripts
        (to avoid multiple scripts responding to a command).
    descr : `str`
        Short description of what the script does, for operator display.

    Raises
    ------
    ValueError
        If index=0. This is prohibited because index=0 would cause
        the script to respond to commands meant for every other script.

    Attributes
    ----------
    log : `logging.Logger`
        A logger.
    done_task : `asyncio.Task`
        A task that is done when the script has fully executed.
    timestamps : `dict` [``lsst.ts.idl.enums.ScriptState``, `float`]
        Dict of script state: TAI unix timestamp.
        Used to set timestamp data in the ``script`` event.
    """

    def __init__(self, index, descr):
        if index == 0:
            raise ValueError("index must be nonzero")

        schema = self.get_schema()
        if schema is None:
            self.config_validator = None
        else:
            self.config_validator = validator.DefaultingValidator(schema=schema)
        self._run_task = None
        self._pause_future = None
        # Value incremented by `next_supplemented_group_id`
        # and cleared by do_setGroupId.
        self._sub_group_id = 0
        self._is_exiting = False
        # Delay (sec) to allow sending the final state and acknowleding
        # the command before exiting.
        self.final_state_delay = 0.3

        # A dict of state: timestamp (TAI seconds).
        self.timestamps = dict()

        self._heartbeat_task = asyncio.Future()

        # Speed up script loading time and avoid expensive system alignments
        # by making sure scripts never become master.
        # The env var must be set while the `DomainParticipant` is created.
        initial_master_prority = os.environ.get(base.MASTER_PRIORITY_ENV_VAR, None)
        os.environ[base.MASTER_PRIORITY_ENV_VAR] = "0"
        try:
            super().__init__("Script", index, do_callbacks=True)
        finally:
            if initial_master_prority is None:
                del os.environ[base.MASTER_PRIORITY_ENV_VAR]
            else:
                os.environ[base.MASTER_PRIORITY_ENV_VAR] = initial_master_prority

        self.evt_description.set(classname=type(self).__name__, description=str(descr))
        self._heartbeat_task = asyncio.create_task(self._heartbeat_loop())

    async def start(self):
        remote_names = set()
        remote_start_tasks = []
        for salinfo in self.domain.salinfo_set:
            if salinfo is self.salinfo:
                continue
            remote_names.add(f"{salinfo.name}:{salinfo.index}")
            remote_start_tasks.append(salinfo.start_task)

        await super().start()
        await asyncio.gather(*remote_start_tasks)

        self.evt_state.set_put(state=ScriptState.UNCONFIGURED)
        self.evt_description.set_put(
            remotes=",".join(sorted(remote_names)), force_output=True
        )

    @classmethod
    def make_from_cmd_line(cls, descr=None):
        """Make a script from command-line arguments.

        Return None if ``--schema`` specified.

        Parameters
        ----------
        descr : `str`, optional
            Short description of what the script does, for operator display.
            Leave at None if the script already has a description, which is
            the most common case. Primarily intended for unit tests,
            e.g. running ``TestScript``.


        Notes
        -----
        The final return code will be:

        * 0 if final state is `lsst.ts.idl.enums.Script.ScriptState.DONE`
          or `lsst.ts.idl.enums.Script.ScriptState.STOPPED`
        * 1 if final state is `lsst.ts.idl.enums.Script.ScriptState.FAILED`
        * 2 otherwise (which should never happen)
        """
        parser = argparse.ArgumentParser(f"Run {cls.__name__} from the command line")
        parser.add_argument(
            "index",
            type=int,
            help="Script SAL Component index; must be unique among running Scripts",
        )
        parser.add_argument(
            "--schema",
            action="store_true",
            help="Print the configuration schema to stdout and quit "
            "without running the script. "
            "The index argument is ignored, though it is still required.",
        )
        args = parser.parse_args()
        kwargs = dict(index=args.index)
        if descr is not None:
            kwargs["descr"] = descr
        if args.schema:
            schema = cls.get_schema()
            if schema is not None:
                print(yaml.safe_dump(schema))
            return
        return cls(**kwargs)

    @classmethod
    async def amain(cls, descr=None):
        """Run the script from the command line.

        Parameters
        ----------
        descr : `str`, optional
            Short description of what the script does, for operator display.
            Leave at None if the script already has a description, which is
            the most common case. Primarily intended for unit tests,
            e.g. running ``TestScript``.


        Notes
        -----
        The final return code will be:

        * 0 if final state is `lsst.ts.idl.enums.Script.ScriptState.DONE`
          or `lsst.ts.idl.enums.Script.ScriptState.STOPPED`
        * 1 if final state is `lsst.ts.idl.enums.Script.ScriptState.FAILED`
        * 2 otherwise (which should never happen)
        """
        script = cls.make_from_cmd_line(descr=descr)
        if script is None:
            # printed schema and exited
            return
        await script.done_task
        return_code = {
            ScriptState.DONE: 0,
            ScriptState.STOPPED: 0,
            ScriptState.FAILED: 1,
        }.get(script.state.state, 2)
        sys.exit(return_code)

    @property
    def checkpoints(self):
        """Get the checkpoints at which to pause and stop.

        Returns ``self.evt_checkpoints.data`` which has these fields:

        * ``pause``: checkpoints at which to pause, a regular expression
        * ``stop``: checkpoints at which to stop, a regular expression
        """
        return self.evt_checkpoints.data

    @property
    def group_id(self):
        """Get the group ID (a `str`), or "" if not set."""
        return self.evt_state.data.groupId

    @property
    def state(self):
        """Get the current state.

        Returns ``self.evt_state.data``, which has these fields:

        * ``state``: `lsst.ts.idl.enums.Script.ScriptState`
            The current state.
        * ``last_checkpoint``: `str`
            Name of most recently seen checkpoint.
        * ``reason``: `str`
            Reason for this state, if any.
        """
        return self.evt_state.data

    @property
    def state_name(self):
        """Get the name of the current `state`.state."""
        try:
            return ScriptState(self.state.state).name
        except ValueError:
            return f"UNKNOWN({self.state.state})"

    def set_state(
        self,
        state=None,
        reason=None,
        keep_old_reason=False,
        last_checkpoint=None,
        force_output=False,
    ):
        """Set the script state.

        Parameters
        ----------
        state : `ScriptState` or `int`, optional
            New state, or None if no change
        reason : `str`, optional
            Reason for state change. `None` for no new reason.
        keep_old_reason : `bool`
            If True, keep old reason; append the ``reason`` argument after ";"
            if it is is a non-empty string.
            If False replace with ``reason``, or "" if ``reason`` is `None`.
        last_checkpoint : `str`, optional
            Name of most recently seen checkpoint. None for no change.
        force_output : `bool`, optional
            If True the output even if not changed.
        """
        if state is not None:
            state = ScriptState(state)
            self.timestamps[state] = base.current_tai()
        if keep_old_reason and reason is not None:
            sepstr = "; " if self.evt_state.data.reason else ""
            reason = self.evt_state.data.reason + sepstr + reason
        self.evt_state.set_put(
            state=state,
            reason=reason,
            lastCheckpoint=last_checkpoint,
            force_output=force_output,
        )

    async def checkpoint(self, name=""):
        """Await this at any "nice" point your script can be paused or stopped.

        Parameters
        ----------
        name : `str`, optional
            Name of checkpoint; "" if it has no name.

        Raises
        ------
        RuntimeError
            If the state is not `ScriptState.RUNNING`. This likely means
            you called checkpoint from somewhere other than `run`.
        RuntimeError
            If `_run_task` is `None` or done. This probably means your code
            incorrectly set the state.
        """
        if not self.state.state == ScriptState.RUNNING:
            raise RuntimeError(
                f"checkpoint error: state={self.state_name} instead of RUNNING; "
                "did you call checkpoint from somewhere other than `run`?"
            )
        if self._run_task is None:
            raise RuntimeError("checkpoint error: state is RUNNING but no run_task")
        if self._run_task.done():
            raise RuntimeError(
                "checkpoint error: state is RUNNING but run_task is done"
            )

        if re.fullmatch(self.checkpoints.stop, name):
            self.set_state(ScriptState.STOPPING, last_checkpoint=name)
            raise asyncio.CancelledError(
                f"stop by request: checkpoint {name} matches {self.checkpoints.stop}"
            )
        elif re.fullmatch(self.checkpoints.pause, name):
            self._pause_future = asyncio.Future()
            self.set_state(ScriptState.PAUSED, last_checkpoint=name)
            await self._pause_future
            self.set_state(ScriptState.RUNNING)
        else:
            self.set_state(last_checkpoint=name, force_output=True)
            await asyncio.sleep(0.001)

    async def close_tasks(self):
        self._is_exiting = True
        await super().close_tasks()
        self._heartbeat_task.cancel()
        if self._run_task is not None:
            self._run_task.cancel()
        if self._pause_future is not None:
            self._pause_future.cancel()
        # self.done_task is handled by Controller.close

    @abc.abstractmethod
    async def configure(self, config):
        """Configure the script.

        Parameters
        ----------
        config : `types.SimpleNamespace`
            Configuration.

        Notes
        -----
        This method is called by `do_configure``.
        The script state will be `ScriptState.UNCONFIGURED`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_metadata(self, metadata):
        """Set metadata fields in the provided struct, given the
        current configuration.

        Parameters
        ----------
        metadata : ``self.evt_metadata.DataType()``
            Metadata to update. Set those fields for which
            you have useful information.

        Notes
        -----
        This method is called after `configure` by `do_configure`.
        The script state will be `ScriptState.UNCONFIGURED`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def run(self):
        """Run the script.

        Your subclass must provide an implementation, as follows:

        * At points where you support pausing call `checkpoint`.
        * Raise an exception on error. Raise `base.ExpectedError`
          to avoid logging a traceback.

        Notes
        -----
        This method is only called when the script state is
        `ScriptState.CONFIGURED`. The remaining state transitions
        are handled automatically.
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_schema(cls):
        """Return a jsonschema to validate configuration, as a `dict`.

        Please provide default values for all fields for which defaults
        make sense. This makes the script easier to use.

        If your script has no configuration then return `None`,
        in which case the ``config`` field of the ``configure`` command
        must be an empty string.
        """
        raise NotImplementedError()

    async def cleanup(self):
        """Perform final cleanup, if any.

        This method is always called as the script state is exiting
        (unless the script process is aborted by SIGTERM or SIGKILL).
        """
        pass

    def assert_state(self, action, states):
        """Assert that the current state is in ``states`` and the script
        is not exiting.

        Parameters
        ----------
        action : `string`
            Description of what you want to do.
        states : `list` [`lsst.ts.idl.enums.Script.ScriptState`]
            Allowed states.
        """
        if self._is_exiting:
            raise base.ExpectedError(f"Cannot {action}: script is exiting")
        if self.state.state not in states:
            states_str = ", ".join(s.name for s in states)
            raise base.ExpectedError(
                f"Cannot {action}: state={self.state_name} instead of {states_str}"
            )

    async def do_configure(self, data):
        """Configure the currently loaded script.

        Parameters
        ----------
        data : ``cmd_configure.DataType``
            Configuration.

        Raises
        ------
        base.ExpectedError
            If ``self.state.state`` is not
            `lsst.ts.idl.enums.Script.ScriptState.UNCONFIGURED`.

        Notes
        -----
        This method does the following:

        * Parse the ``config`` field as yaml-encoded `dict` and validate it
          (including setting default values).
        * Call `configure`.
        * Set the pause and stop checkpoints.
        * Set the log level if ``data.logLevel != 0``.
        * Call `set_metadata`.
        * Output the metadata event.
        * Change the script state to
          `lsst.ts.idl.enums.Script.ScriptState.CONFIGURED`.
        """
        self.assert_state("configure", [ScriptState.UNCONFIGURED])
        try:
            if self.config_validator is None:
                if data.config:
                    raise RuntimeError(
                        "This script has no configuration; "
                        f"config={data.config} must be empty."
                    )
                config = types.SimpleNamespace()
            else:
                if data.config:
                    user_config_dict = yaml.safe_load(data.config)
                    # Delete metadata, if present
                    if user_config_dict:
                        user_config_dict.pop("metadata", None)
                else:
                    user_config_dict = {}
                full_config_dict = self.config_validator.validate(user_config_dict)
                config = types.SimpleNamespace(**full_config_dict)
            await self.configure(config)
        except Exception as e:
            errmsg = f"config({data.config}) failed"
            self.log.exception(errmsg)
            raise base.ExpectedError(f"{errmsg}: {e}") from e

        self._set_checkpoints(pause=data.pauseCheckpoint, stop=data.stopCheckpoint)
        if data.logLevel != 0:
            self.log.setLevel(data.logLevel)
            self.put_log_level()

        metadata = self.evt_metadata.DataType()
        # initialize to vaguely reasonable values
        metadata.coordinateSystem = MetadataCoordSys.NONE
        metadata.rotationSystem = MetadataRotSys.NONE
        metadata.filters = ""  # any
        metadata.dome = MetadataDome.EITHER
        metadata.duration = 0
        self.set_metadata(metadata)
        self.evt_metadata.put(metadata)
        self.set_state(ScriptState.CONFIGURED)
        await asyncio.sleep(0.001)

    async def do_run(self, data):
        """Run the script and quit.

        The script must have been configured and the group ID set.

        Parameters
        ----------
        data : ``cmd_run.DataType``
            Ignored.

        Raises
        ------
        base.ExpectedError
            If ``self.state.state`` is not
            `lsst.ts.idl.enums.Script.ScriptState.CONFIGURED`.
            If ``self.group_id`` is blank.
        """
        self.assert_state("run", [ScriptState.CONFIGURED])
        if not self.group_id:
            raise base.ExpectedError("Group ID not set")
        try:
            self.set_state(ScriptState.RUNNING)
            self._run_task = asyncio.create_task(self.run())
            await self._run_task
            self.set_state(ScriptState.ENDING)
        except asyncio.CancelledError:
            if self.state.state != ScriptState.STOPPING:
                self.set_state(ScriptState.STOPPING)
        except Exception as e:
            if not isinstance(e, base.ExpectedError):
                self.log.exception("Error in run")
            self.set_state(ScriptState.FAILING, reason=f"Error in run: {e}")
        await asyncio.sleep(0.001)
        await self._exit()

    async def do_resume(self, data):
        """Resume the currently paused script.

        Parameters
        ----------
        data : ``cmd_resume.DataType``
            Ignored.

        Raises
        ------
        base.ExpectedError
            If ``self.state.state`` is not
            `lsst.ts.idl.enums.Script.ScriptState.PAUSED`.
        """
        self.assert_state("resume", [ScriptState.PAUSED])
        self._pause_future.set_result(None)

    async def do_setCheckpoints(self, data):
        """Set or clear the checkpoints at which to pause and stop.

        This command is deprecated. Please set the checkpoints
        using the `configure` command.

        Parameters
        ----------
        data : ``cmd_setCheckpoints.DataType``
            Names of checkpoints for pausing and stopping, each a single
            regular expression; "" for no checkpoints, ".*" for all.

        Raises
        ------
        base.ExpectedError
            If ``self.state.state`` is not one of:

            * `lsst.ts.idl.enums.Script.ScriptState.UNCONFIGURED`
            * `lsst.ts.idl.enums.Script.ScriptState.CONFIGURED`
            * `lsst.ts.idl.enums.Script.ScriptState.RUNNING`
            * `lsst.ts.idl.enums.Script.ScriptState.PAUSED`
        """
        self.assert_state(
            "setCheckpoints",
            [
                ScriptState.UNCONFIGURED,
                ScriptState.CONFIGURED,
                ScriptState.RUNNING,
                ScriptState.PAUSED,
            ],
        )
        self._set_checkpoints(pause=data.pause, stop=data.stop)

    async def do_setGroupId(self, data):
        """Set or clear the group_id attribute.

        The script must be in the Configured state.
        This command may be called multiple times. It is typically called
        when the script reaches the top position on the script queue.

        Parameters
        ----------
        data : ``cmd_setGroupId.DataType``
            Group ID, or "" to clear the group ID.

        Raises
        ------
        base.ExpectedError
            If ``state.state`` is not
            `lsst.ts.idl.enums.Script.ScriptState.CONFIGURED`.
        """
        self.assert_state("setGroupId", [ScriptState.CONFIGURED])
        self.evt_state.set_put(groupId=data.groupId, force_output=True)
        self._sub_group_id = 0

    async def do_stop(self, data):
        """Stop the script.

        Parameters
        ----------
        data : ``cmd_stop.DataType``
            Ignored.

        Notes
        -----
        This is a no-op if the script is already exiting.
        This does not wait for _exit to run.
        """
        if self._is_exiting:
            return
        if self._run_task is not None and not self._run_task.done():
            self._run_task.cancel()
        else:
            self.set_state(state=ScriptState.STOPPING)
            await self._exit()

    def next_supplemented_group_id(self):
        """Return the group ID supplemented with a new subgroup.

        The returned string has this format: f"{self.group_id}#{subgroup_id}",
        where ``subgroup_id`` is an integer that starts at 1
        and is incremented for every call to this method.

        Raises
        ------
        RuntimeError
            If there is no group ID.
        """
        if not self.group_id:
            raise RuntimeError("No group ID")
        self._sub_group_id += 1
        return f"{self.group_id}#{self._sub_group_id}"

    def _set_checkpoints(self, *, pause, stop):
        """Set the pause and stop checkpoint fields and output the event.

        Parameters
        ----------
        pause : `string`
            Checkpoint(s) at which to pause, as a regular expression.
            "" to not pause at any checkpoint; "*" to pause at all checkpoints.
        stop : `string`
            Checkpoint(s) at which to stop, as a regular expression.
            "" to not stop at any checkpoint; "*" to stop at all checkpoints.

        Raises
        ------
        lsst.ts.base.ExpectedError
            If pause or stop are not valid regular expressions.
        """
        try:
            re.compile(pause)
        except Exception as e:
            raise base.ExpectedError(f"pause={pause!r} not a valid regex: {e}")
        try:
            re.compile(stop)
        except Exception as e:
            raise base.ExpectedError(f"stop={stop!r} not a valid regex: {e}")
        self.evt_checkpoints.set_put(pause=pause, stop=stop, force_output=True)

    async def _heartbeat_loop(self):
        """Output heartbeat at regular intervals."""
        while True:
            try:
                await asyncio.sleep(HEARTBEAT_INTERVAL)
                self.evt_heartbeat.put()
            except asyncio.CancelledError:
                break
            except Exception:
                self.log.exception("Heartbeat output failed")

    async def _exit(self):
        """Call cleanup (if the script was run) and exit the script."""
        if self._is_exiting:
            return
        self._is_exiting = True
        try:
            if self._run_task is not None:
                await self.cleanup()
            self._heartbeat_task.cancel()

            reason = None
            final_state = {
                ScriptState.ENDING: ScriptState.DONE,
                ScriptState.STOPPING: ScriptState.STOPPED,
                ScriptState.FAILING: ScriptState.FAILED,
            }.get(self.state.state)
            if final_state is None:
                reason = f"unexpected state for _exit {self.state_name}"
                final_state = ScriptState.FAILED

            self.log.info(f"Setting final state to {final_state!r}")
            self.set_state(final_state, reason=reason, keep_old_reason=True)
            asyncio.create_task(self.close())
        except Exception as e:
            if not isinstance(e, base.ExpectedError):
                self.log.exception("Error in run")
            self.set_state(
                ScriptState.FAILED, reason=f"failed in _exit: {e}", keep_old_reason=True
            )
            asyncio.create_task(self.close(exception=e))



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

import numpy as np

from lsst.ts import salobj

np.random.seed(47)


class BasicCscCommander(salobj.TestCscCommander):
    """A version of `TestCscCommander` that adds a custom "echo" command.

    This class adds a custom command ("echo") that is not supported by
    the Test CSC. This includes help for the command.

    Inherit from `TestCscCommander` instead of `CscCommander` in order to
    test a custom override of a command supported by the Test CSC
    ("setArrays") and to test hiding unsupported generic commands.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.help_dict["echo"] = "any  # echo arguments, space-separated"

    def do_echo(self, args):
        """Output the arguments."""
        self.output(" ".join(args))


class CscCommanderTestCase(salobj.BaseCscTestCase, unittest.IsolatedAsyncioTestCase):
    def basic_make_csc(self, initial_state, config_dir, simulation_mode):
        index = self.next_index()
        self.commander = BasicCscCommander(index=index)
        self.addAsyncCleanup(self.commander.close)
        self.commander.testing = True

        return salobj.TestCsc(
            index=index,
            initial_state=initial_state,
            config_dir=config_dir,
            simulation_mode=simulation_mode,
        )

    async def test_basics(self):
        async with self.make_csc(initial_state=salobj.State.STANDBY):
            await self.commander.start()
            await self.assert_next_summary_state(salobj.State.STANDBY)

            # Test a command that will fail because the CSC is not enabled
            with salobj.assertRaisesAckError(ack=salobj.SalRetCode.CMD_FAILED):
                await self.commander.run_command(
                    f"wait {salobj.SalRetCode.CMD_COMPLETE} 5"
                )
            # Test some standard CSC commands
            await self.commander.run_command("start")
            await self.assert_next_summary_state(salobj.State.DISABLED)
            await self.commander.run_command("enable")
            await self.assert_next_summary_state(salobj.State.ENABLED)
            t0 = salobj.current_tai()
            wait_time = 2  # seconds
            await self.commander.run_command(
                f"wait {salobj.SalRetCode.CMD_COMPLETE} {wait_time}"
            )
            dt = salobj.current_tai() - t0
            # The margin of 0.2 compensates for the clock in Docker on macOS
            # not being strictly monotonic.
            self.assertGreaterEqual(dt, wait_time - 0.2)
            self.commander.output_queue.clear()

            # Test the help command
            await self.commander.run_command("help")
            self.assertEqual(len(self.commander.output_queue), 1)
            help_str = self.commander.output_queue.pop()
            for substr in ("standby", "exitControl", "echo", "help"):
                self.assertIn(substr, help_str)
            for hidden_command in ("abort", "enterControl", "setValue"):
                self.assertNotIn(hidden_command, help_str)

            # Test TestCscCommander's custom setArrays command handler
            # Too many values
            with self.assertRaises(RuntimeError):
                await self.commander.run_command("setArrays int0=1,2,3,4,5,6")
            # No such field
            with self.assertRaises(RuntimeError):
                await self.commander.run_command("setArrays no_such_field=1,2,3,4,5,6")
            # A valid command
            await self.commander.run_command("setArrays boolean0=0,1 int0=-2,33,42")
            for topic in self.csc.evt_arrays, self.csc.tel_arrays:
                self.assertEqual(
                    topic.data.boolean0, [False, True, False, False, False]
                )
                self.assertEqual(topic.data.int0, [-2, 33, 42, 0, 0])

            # Test bool argument handling. Set all other fields to 0
            # and ignore them.
            scalar_fields = [
                field
                for field in self.csc.cmd_setScalars.DataType().get_vars().keys()
                if not field.startswith("private_") and field != "TestID"
            ]
            bool_index = scalar_fields.index("boolean0")
            n_scalar_fields = len(scalar_fields)
            for good_bool_arg, value in (
                ("0", False),
                ("f", False),
                ("F", False),
                ("FaLsE", False),
                ("1", True),
                ("t", True),
                ("T", True),
                ("tRuE", True),
            ):
                args = ["0"] * n_scalar_fields
                args[bool_index] = good_bool_arg
                await self.commander.run_command("setScalars " + " ".join(args))
                self.assertEqual(self.csc.evt_scalars.data.boolean0, value)

            for bad_bool_arg in ("2", "fail", "falsely"):
                args = ["0"] * n_scalar_fields
                args[bool_index] = bad_bool_arg
                with self.assertRaises(ValueError):
                    await self.commander.run_command("setScalars " + " ".join(args))

            # Test BasicCscCommander's "echo" command
            self.commander.output_queue.clear()
            argstr = "1   2.3   arbitrary   text"
            await self.commander.run_command(f"echo {argstr}")
            self.assertEqual(len(self.commander.output_queue), 1)
            expected_output = " ".join(argstr.split())
            output_str = self.commander.output_queue.pop()
            self.assertEqual(output_str, expected_output)


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["ControllerEvent"]

from . import write_topic


class ControllerEvent(write_topic.WriteTopic):
    """Write a specific event topic.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Event topic name
    """

    def __init__(self, salinfo, name):
        super().__init__(salinfo=salinfo, name=name, sal_prefix="logevent_")

    def set_put(self, *, force_output=False, **kwargs):
        """Set zero or more fields of ``self.data`` and put if changed
        or if ``force_output`` true.

        The data is put if it has never been set (`has_data` False), or this
        call changes the value of any field, or ``force_output`` is true.

        Parameters
        ----------
        force_output : `bool`, optional
            If True then output the event, even if no fields have changed.
        **kwargs : `dict` [`str`, ``any``]
            The remaining keyword arguments are
            field name = new value for that field.
            See `set` for more information about values.

        Returns
        -------
        did_put : `bool`
            True if the data was output, False otherwise

        Raises
        ------
        AttributeError
            If the topic does not have the specified field.
        ValueError
            If the field cannot be set to the specified value.
        """
        did_change = self.set(**kwargs)
        do_output = did_change or force_output
        if do_output:
            self.put(self.data)
        return do_output



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import logging
import pathlib
import time
import unittest
import warnings

from lsst.ts import salobj

try:
    import SALPY_Test
except ImportError:
    SALPY_Test = None

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

INITIAL_LOG_LEVEL = 20
SAL__CMD_COMPLETE = 303

index_gen = salobj.index_generator()


class SALPYTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        self.datadir = pathlib.Path(__file__).resolve().parent / "data"
        self.index = next(index_gen)

    async def test_salobj_remote_salobj_controller(self):
        await self.check_salobj_remote("minimal_salobj_controller.py")

    @unittest.skipIf(SALPY_Test is None, "Could not import SALPY_Test")
    async def test_salobj_remote_salpy_controller(self):
        await self.check_salobj_remote("minimal_salpy_controller.py")

    async def check_salobj_remote(self, exec_name):
        # Create the remote before the subprocess
        # to be sure the remote sees telemetry from the subprocess
        # (telemetry is volatile, so has no historical data).
        print(f"Remote: create salobj remote with index={self.index}")
        t0 = time.monotonic()
        async with salobj.Domain() as domain:
            # ts_sal requires a CSC identity
            # (or explicitly issuing the setAuthList command)
            domain.default_identity = f"Test:{self.index}"
            async with salobj.Remote(
                domain=domain,
                name="Test",
                index=self.index,
                evt_max_history=1,
            ) as remote:
                dt = time.monotonic() - t0
                print(
                    f"Remote: creating topics and waiting for historical data took {dt:0.2f} seconds"
                )
                remote.salinfo.log.addHandler(logging.StreamHandler())

                print(f"Remote: start {exec_name} in a subprocess")
                script_path = self.datadir / exec_name
                process = await asyncio.create_subprocess_exec(
                    str(script_path), str(self.index), str(INITIAL_LOG_LEVEL)
                )

                try:
                    print("Remote: wait for initial logLevel event")
                    data = await remote.evt_logLevel.next(
                        flush=False, timeout=STD_TIMEOUT
                    )
                    print(f"Remote: read initial logLevel.level={data.level}")
                    self.assertEqual(data.level, INITIAL_LOG_LEVEL)

                    for level in (10, 52, 0):
                        # remote.cmd_setLogLevel.put()
                        # print(f"Remote: put setLogLevel(level={level})")
                        print(f"Remote: send setLogLevel(level={level}) command")
                        ackcmd = await remote.cmd_setLogLevel.set_start(
                            level=level, timeout=STD_TIMEOUT
                        )
                        self.assertEqual(ackcmd.identity, remote.salinfo.identity)
                        print("Remote: wait for logLevel event")
                        data = await remote.evt_logLevel.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        print(f"Remote: read logLevel={data.level}")
                        self.assertEqual(data.level, level)
                        print("Remote: wait for scalars telemetry")
                        data = await remote.tel_scalars.next(
                            flush=False, timeout=STD_TIMEOUT
                        )
                        print(f"Remote: read scalars.int0={data.int0}")
                        self.assertEqual(data.int0, level)
                        await asyncio.sleep(0.1)

                    await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
                finally:
                    print("Remote: done")
                    if process.returncode is None:
                        process.terminate()
                        warnings.warn(
                            "Killed a process that was not properly terminated"
                        )


if __name__ == "__main__":
    unittest.main()



import enum

__all__ = ["SalRetCode", "State", "as_salRetCode", "as_state"]


class SalRetCode(enum.IntEnum):
    """SAL return codes."""

    OK = 0
    ERR = -1
    ERROR = -1
    ILLEGAL_REVCODE = -2
    TOO_MANY_HANDLES = -3
    NOT_DEFINED = -4
    # Generic State machine states
    STATE_DISABLED = 1
    STATE_ENABLED = 2
    STATE_FAULT = 3
    STATE_OFFLINE = 4
    STATE_STANDBY = 5
    # Timeout return codes
    TIMEOUT = -5
    SIGNAL_INTERRUPT = -6
    # getSample timeout specifiers (+ve is a time in microseconds)
    WAIT_FOR_NEXT_UPDATE = -10000
    WAIT_FOR_CHANGE = -10001
    # telemetry stream update types
    NO_UPDATES = -100
    WAITING_FOR_NEXT = 100
    GOT_UPDATE = 101
    SYNC_IN = 102
    SYNC_OUT = 103
    SYNC_SET = 104
    SYNC_CLEAR = 105
    SYNC_READ = 106
    # generateAlert types
    EVENT_INFO = 200
    EVENT_WARN = -200
    EVENT_ERROR = -201
    EVENT_ABORT = -202
    # issueCommand/getResponse return codes
    CMD_ACK = 300
    CMD_INPROGRESS = 301
    CMD_STALLED = 302
    CMD_COMPLETE = 303
    CMD_NOPERM = -300
    CMD_NOACK = -301
    CMD_FAILED = -302
    CMD_ABORTED = -303
    CMD_TIMEOUT = -304
    # callback types for subscriptions
    DATA_AVAIL = 400
    DEADLINE_MISS = 401
    INCOMPAT_QOS = 402
    SAMPLE_REJ = 403
    LIVELINESS_CHG = 404
    SAMPLELOST = 405
    SUBSCR_MATCH = 406


class State(enum.IntEnum):
    """CSC summaryState constants."""

    OFFLINE = 4
    STANDBY = 5
    DISABLED = 1
    ENABLED = 2
    FAULT = 3


def as_salRetCode(value):
    """Convert an int (or SalRetCode) to a SalRetCode.

    Return the original value if no match.
    """
    try:
        return SalRetCode(value)
    except Exception:
        return value


def as_state(value):
    """Convert an int (or State) to a State.

    Return the original value if no match.
    """
    try:
        return State(value)
    except Exception:
        return value



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import pathlib
import unittest

from lsst.ts import salobj


class ConfigTestCase(salobj.BaseConfigTestCase, unittest.TestCase):
    def setUp(self):
        self.schema = salobj.CONFIG_SCHEMA

    def test_get_schema(self):
        csc_package_root = pathlib.Path(__file__).parents[1]
        schema = self.get_schema(csc_package_root=csc_package_root, sal_name="Test")
        self.assertIsInstance(schema, dict)

        schema2 = self.get_schema(
            csc_package_root=csc_package_root, schema_subpath="schema/Test.yaml"
        )
        self.assertEqual(schema, schema2)

        with self.assertRaises(AssertionError):
            # Invalid sal_name
            self.get_schema(
                csc_package_root=csc_package_root, sal_name="NoSuchSalComponent"
            )
        with self.assertRaises(AssertionError):
            # Invalid schema_subpath
            self.get_schema(
                csc_package_root=csc_package_root,
                schema_subpath="schema/no_such_file.yaml",
            )
        with self.assertRaises(AssertionError):
            # Invalid csc_package_root
            self.get_schema(csc_package_root="not/a/directory", sal_name="Test")
        with self.assertRaises(RuntimeError):
            # Must specify sal_name or schema_subpath
            self.get_schema(csc_package_root=csc_package_root)

    def test_get_module_dir(self):
        module_dir = self.get_module_dir("lsst.ts.salobj")
        self.assertTrue(str(module_dir).endswith("lsst/ts/salobj"))

        with self.assertRaises(ModuleNotFoundError):
            self.get_module_dir("lsst.lsst.lsst.no_such_module")

    @unittest.skipIf("TS_CONFIG_OCS_DIR" not in os.environ, "ts_config_ocs not found")
    def test_standard_configs(self):
        """Test the config files in ts_config_ocs/Test/..."""
        config_package_root = os.environ["TS_CONFIG_OCS_DIR"]
        config_dir = self.get_config_dir(
            config_package_root=config_package_root, sal_name="Test", schema=self.schema
        )
        self.check_config_files(config_dir=config_dir, schema=self.schema)

        # Test check_standard_config_files using module import and schema_name
        self.check_standard_config_files(
            module_name="lsst.ts.salobj",
            schema_name="CONFIG_SCHEMA",
            config_dir=config_dir,
        )

        # Test check_standard_config_files using module import
        self.check_standard_config_files(
            sal_name="Test", module_name="lsst.ts.salobj", config_dir=config_dir
        )

        # Test check_standard_config_files using env var TS_SALOBJ_DIR
        self.check_standard_config_files(
            sal_name="Test", package_name="ts_salobj", config_dir=config_dir
        )

        # Test check_standard_config_files using env var TS_SALOBJ_DIR
        # and schema_subpath
        self.check_standard_config_files(
            sal_name="Test",
            package_name="ts_salobj",
            config_dir=config_dir,
            schema_subpath="schema/Test.yaml",
        )

        with self.assertRaises(AssertionError):
            self.check_standard_config_files(
                sal_name="Test",
                package_name="ts_salobj",
                config_dir=config_dir,
                schema_subpath="schema/no_such_file.yaml",
            )

    def test_local_configs(self):
        """Test the various local config directories."""
        data_root = pathlib.Path(__file__).parent / "data"

        self.check_config_files(
            config_dir=data_root / "config_good", schema=self.schema
        )

        for config_dir in data_root.glob("config_bad*"):
            with self.subTest(config_dir=str(config_dir)):
                with self.assertRaises(AssertionError):
                    self.check_config_files(config_dir=config_dir, schema=self.schema)


if __name__ == "__main__":
    unittest.main()
